# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.
#
#
# The following python code creates small .app5 files for unit testing by
# downsampling larger experimental datasets from various labs around the
# world (Details below).

# This file will not work as-written without the original files, but it is
# still provided for two reasons:

#     1) Transparency and validation.
#     2) To create additional test files if future versions of TopSpin
#        are incompatible with this reader.

# Additionally, while the datasets themselves have a consistent layout,
# their location within the .app5 files changes based on how the
# file was exported from TopSpin.


import os
import shutil
import subprocess
import xml.etree.ElementTree as ET
from pathlib import Path

import h5py
import numpy as np


def downsample_4D_group(group, mask):
    """In-place reduction of hdf5 group representing 4D-STEM data."""
    pxls = mask.flatten() * np.arange(mask.size)
    pxls_dict = dict([str(x), str(i)] for i, x in enumerate(pxls[pxls > 0]))
    gnames = [str(x) for x in np.arange(len(group.keys()))]
    for gname in gnames:
        if gname in pxls_dict.keys():
            new_id = pxls_dict[gname]
            group.move(gname, new_id)
            data = group[new_id]["Data"][::19, ::23].copy()
            del group[new_id]["Data"]
            group[new_id].create_dataset("Data", dtype="<f4", data=data)
        else:
            del group[gname]


def downsample_2D_dataset(group, guid):
    """In-place reduction of 2D stem image."""
    img = group[guid][::7, ::9]
    del group[guid]
    group.create_dataset(guid, dtype="<f8", data=img)


out_dname = Path(__file__).parent / "data" / "topspin"

# %% Test file 1: Colorado School of Mines single-run file.
# hdf5 layout:
# | cb76b58d-7906-423d-b681-3af9587e9e76.app5
# | - '39a4bbcd-2551-4912-9093-0eaed678dcff'  <-- STEM dataset
# | - 'Metadata'
# | - 'a0dba950-15bf-4a74-a81c-9cea8560c2c8' <-- 4DSTEM group
# | --- '1'
# | ----- 'Data'
# | ----- 'Metadata'
# | --- ....
# | - 'Version'

# Copy 6.6 GB app5 file
old_fname = "/home/arg6/data/app5/cb76b58d-7906-423d-b681-3af9587e9e76.app5"
temp_fname = "A.temp"
final_fname = out_dname / "topspin_test_A.app5"
if os.path.exists(temp_fname):
    os.remove(temp_fname)
if os.path.exists(final_fname):
    os.remove(final_fname)
shutil.copyfile(old_fname, temp_fname)
app5_A = h5py.File(temp_fname, "a")

# downsample '39a4bbcd-2551-4912-9093-0eaed678dcff' (2D virtual STEM image).
downsample_2D_dataset(app5_A, "39a4bbcd-2551-4912-9093-0eaed678dcff")

# downsample 'a0dba950-15bf-4a74-a81c-9cea8560c2c8' (4D STEM group)
mask = np.zeros([50, 500], dtype=bool)
mask[10:12, 50:55] = True
downsample_4D_group(app5_A["a0dba950-15bf-4a74-a81c-9cea8560c2c8"], mask)

# Edit toplevel metadata to better reflect contributions.
# Done manually as different TopSpin setups store MetaData slightly different.
root = ET.fromstring(app5_A["Metadata"][()].decode())
root[7].text = "Test data A for rsciio"
root[12].text = "Edwin Supple"
root[13].text = (
    "Collected at the Colorado School of Mines for NIST's Quantitative"
    + " Nanostructure Characterization Group."
)
txt = ET.tostring(root, xml_declaration=True)
app5_A.move("Metadata", "temp")
app5_A.create_dataset_like("Metadata", app5_A["temp"], data=txt)
del app5_A["temp"]

# repackage to remove empty space
app5_A.close()
subprocess.run(["h5repack", temp_fname, final_fname])
subprocess.run(["rm", temp_fname])


# %% Test file 2: Colorado School of Mines exported via Topspin.
# hdf5 layout:
# | nanomegas_topspin_test_file.app5
# |
# | - '18d9446f-22bf-4fb1-8d13-338174e75d20'   <-- Session A
# | --- '3526f008-a687-41fb-a21e-c21362241492' <-- 4DSTEM group
# | ----- '1'
# | ------- 'Data'
# | ------- 'Metadata'   <-- Metadata for single pixel image
# | ----- ....
# | --- 'Metadata' <-- Metadata for session
# | --- 'Version'
# | --- 'b0212ada-90b1-4117-8ed5-355717e4910b' <-- STEM dataset
# | --- 'b0e00442-4f83-4585-b84c-b82d2537b14b' <-- STEM dataset
# |
# | - 'b38446c6-6b25-4fee-9c62-6d8e4fcbeb5c'   <-- Session B
# | --- ....
# |
# | - 'Version'


# Copy 1.7 GB app5 file.
old_fname = "/home/arg6/data/app5/1750 1.app5"
temp_fname = "B.app5"
final_fname = out_dname / "topspin_test_B.app5"
if os.path.exists(temp_fname):
    os.remove(temp_fname)
if os.path.exists(final_fname):
    os.remove(final_fname)
shutil.copyfile(old_fname, temp_fname)
app5_B = h5py.File(temp_fname, "a")

# session 1
B1 = app5_B["18d9446f-22bf-4fb1-8d13-338174e75d20"]
# Downsample '3526f008-a687-41fb-a21e-c21362241492' (4D STEM group)
mask = np.zeros([50, 50], dtype=bool)
mask[5:8, 4:11] = True
downsample_4D_group(B1["3526f008-a687-41fb-a21e-c21362241492"], mask)
# Downsample both virtual stem images (2D virtual STEM image)
for guid in [
    "b0212ada-90b1-4117-8ed5-355717e4910b",
    "b0e00442-4f83-4585-b84c-b82d2537b14b",
]:
    downsample_2D_dataset(B1, guid)

# session 2
B2 = app5_B["b38446c6-6b25-4fee-9c62-6d8e4fcbeb5c"]
# Copy "5a203dd8-ddbd-4775-8786-78d456e8b877" (2D virtual STEM image)
downsample_2D_dataset(B2, "5a203dd8-ddbd-4775-8786-78d456e8b877")

# Edit toplevel metadata to better reflect contributions.
for grp in [B1, B2]:
    root = ET.fromstring(grp["Metadata"][()].decode())
    root[7].text = "Test data B for rsciio"
    root[12].text = "Edwin Supple"
    root[13].text = (
        "Collected at the Colorado School of Mines for NIST's Quantitative"
        + " Nanostructure Characterization Group."
    )
    txt = ET.tostring(root, xml_declaration=True)
    grp.move("Metadata", "temp")
    grp.create_dataset_like("Metadata", grp["temp"], data=txt)
    del grp["temp"]

# repackage to remove empty space
app5_B.close()
subprocess.run(["h5repack", temp_fname, final_fname])
subprocess.run(["rm", temp_fname])

# %% Test file 3: Exported from Topspin at University of Glasgow.
#
# hdf5 layout is identical to Test file 2, but Metadata is formatted
# differently.

# copy 2.1Gb app5 file.
old_fname = "/home/arg6/data/app5/IM452.app5"
temp_fname = "C.app5"
final_fname = out_dname / "topspin_test_C.app5"
if os.path.exists(temp_fname):
    os.remove(temp_fname)
if os.path.exists(final_fname):
    os.remove(final_fname)
shutil.copyfile(old_fname, temp_fname)
app5_C = h5py.File(temp_fname, "a")

# session 1
C1 = app5_C["030444c4-694d-4184-94be-4bb9ffeda552"]
# Downsample 'e0c99515-f43c-44c3-8394-132dafbacc39' (4D STEM group)
mask = np.zeros([40, 200], dtype=bool)
mask[2:5, 3:8] = True
downsample_4D_group(C1["e0c99515-f43c-44c3-8394-132dafbacc39"], mask)
# Downsample 'b5b0b75e-a5ff-4cdc-95d6-cab450c49e09' (2D virtual STEM image)
downsample_2D_dataset(C1, "b5b0b75e-a5ff-4cdc-95d6-cab450c49e09")


# session 2
C2 = app5_C["9ebd1693-3de8-4294-877a-62670150b9fa"]
# Copy '8bc2d64f-ea0e-420c-9099-51a07131f6e1' (2D virtual STEM image)
downsample_2D_dataset(C2, "8bc2d64f-ea0e-420c-9099-51a07131f6e1")

# Edit toplevel metadata to better reflect contributions.
for grp in [C1, C2]:
    root = ET.fromstring(grp["Metadata"][()].decode())
    root[7].text = "Test data C for rsciio"
    root[12].text = "Nidhi Choudhary and Ian MacLaren"
    root[13].text = (
        "Collected at the University of Glasgow as part of the following"
        + " paper: 10.1063/5.0292737."
    )
    txt = ET.tostring(root, xml_declaration=True)
    grp.move("Metadata", "temp")
    grp.create_dataset_like("Metadata", grp["temp"], data=txt)
    del grp["temp"]

# repackage to remove empty space
app5_C.close()
subprocess.run(["h5repack", temp_fname, final_fname])
subprocess.run(["rm", temp_fname])



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import ast
import logging
import os
import xml.etree.ElementTree as ET

import numpy as np

from rsciio._docstrings import FILENAME_DOC, LAZY_DOC, RETURNS_DOC
from rsciio.utils import file, xml

_logger = logging.getLogger(__name__)


def _read_raw(info, filename, lazy=False):
    width = info["width"]
    height = info["height"]

    # Make custom dtype to skip the 2 lines footer
    dtype = np.dtype(
        [
            ("data", "<f4", (height, width)),
            ("footer", "<f4", (info["raw_height"] - height, width)),
        ]
    )

    if "series_count" in info.keys():  # stack of images
        navigation_shape = (info["series_count"],)
    else:  # 2D x 2D
        navigation_shape = (info["scan_y"], info["scan_x"])

    if lazy:
        data = file.memmap_distributed(
            filename,
            dtype=dtype,
            shape=navigation_shape,
            key="data",
        )
    else:
        data = np.fromfile(
            filename,
            dtype=dtype,
            count=np.prod(navigation_shape),
        ).reshape(navigation_shape)["data"]

    return data


def _parse_xml(filename):
    tree = ET.parse(filename)
    om = xml.convert_xml_to_dict(tree.getroot())

    info = {
        "raw_filename": om.root.raw_file.filename,
        "width": 128,
        "height": 128,
        "raw_height": 130,
    }
    if om.has_item("root.scan_parameters.series_count"):
        # Stack of images
        info.update({"series_count": int(om.root.scan_parameters.series_count)})
    # The pix_x and pix_y are no longer used in the new EMPAD versions, but keep this line for backward compatibility
    elif om.has_item("root.pix_x") and om.has_item("root.pix_y"):
        # 2D x 2D
        info.update({"scan_x": int(om.root.pix_x), "scan_y": int(om.root.pix_y)})
    # in case root.pix_x and root.pix_y are not available
    elif om.has_item("root.scan_parameters.scan_resolution_x") and om.has_item(
        "root.scan_parameters.scan_resolution_y"
    ):
        info.update(
            {
                "scan_x": int(om.root.scan_parameters.scan_resolution_x),
                "scan_y": int(om.root.scan_parameters.scan_resolution_y),
            }
        )
    else:
        raise IOError("Unsupported Empad file: the scan parameters cannot be imported.")

    return om, info


def _convert_scale_units(value, units, factor=1):
    from rsciio.utils._units import _UREG

    v = float(value) * _UREG(units)
    converted_v = (factor * v).to_compact()
    converted_value = converted_v.magnitude / factor
    converted_units = "{:~}".format(converted_v.units)

    return converted_value, converted_units


def file_reader(filename, lazy=False, q_calibration=None, remove_nans=False):
    """
    Read file format used by the Electron Microscope Pixel Array Detector (EMPAD).

    Parameters
    ----------
    %s
    %s
    q_calibration : None or float, optional
        Specifies the calibration for the diffraction patterns in 1/nm.
        In EMPAD version v1.2.2 (default version shipped with most of the TFS TEMs),
        diffraction space calibration is no longer stored in the XML.
        Use this option to calibrate the diffraction patterns in 1/nm.
        If None, the data will be in pixel scales.
    remove_nans : bool, optional
        Sometimes the EMPAD data contains NaN values that cause issues in
        downstream processing. If True, these NaN values will be replaced
        with zeros. Default is False.

    %s

    Notes
    -----
    For EMPAD file v1.2.0+, the diffraction space calibration is no longer
    stored in the XML. The ``q_calibration`` parameter can be used to set
    the diffraction space calibration in 1/nm.

    Examples
    --------
    >>> from rsciio.empad import file_reader
    >>> s = file_reader("empad_file.xml", q_calibration=0.3315)
    """
    om, info = _parse_xml(filename)
    dname, fname = os.path.split(filename)

    md = {
        "General": {
            "original_filename": fname,
            "title": os.path.splitext(fname)[0],
        },
        "Signal": {"signal_type": "electron_diffraction"},
    }

    if om.has_item("root.timestamp.isoformat"):
        date, time = om.root.timestamp.isoformat.split("T")
        md["General"].update({"date": date, "time": time})

    units = [
        None,
    ] * 2
    scales = [
        1,
    ] * 2
    origins = [
        -64,
    ] * 2
    axes = []
    index_in_array = 0
    names = ["height", "width"]
    navigate = [False, False]

    if "series_count" in info.keys():
        names = ["series_count"] + names
        units.insert(0, "ms")
        scales.insert(0, 1)
        origins.insert(0, 0)
        navigate.insert(0, True)
    else:
        names = ["scan_y", "scan_x"] + names
        units.insert(0, "")
        units.insert(0, "")
        scales.insert(0, 1)
        scales.insert(0, 1)
        origins.insert(0, 0)
        origins.insert(0, 0)
        navigate.insert(0, True)
        navigate.insert(0, True)

    sizes = [info[name] for name in names]

    if "series_count" not in info.keys():
        # Try to read the scan size from the xml
        if om.has_item("root.iom_measurements.optics.get_full_scan_field_of_view"):
            # Keep this for backward compatibility
            fov = ast.literal_eval(
                om.root.iom_measurements.optics.get_full_scan_field_of_view
            )
            for i in range(2):
                value = fov[i] / sizes[i]
                scales[i], units[i] = _convert_scale_units(value, "m", sizes[i])

        elif om.has_item(
            "root.iom_measurements.full_scan_field_of_view.x"
        ) and om.has_item("root.iom_measurements.full_scan_field_of_view.scale_factor"):
            # The fov is stored under iom_measurements.full_scan_field_of_view.x and y,
            # but x and y are always the same, representing the larger dimension of the scan
            fov = ast.literal_eval(om.root.iom_measurements.full_scan_field_of_view.x)
            # The scale factor is to match the EMPAD fov to the internal scan, usually 0.72.
            # The fov is mistakenly stored as the internal fov * scale_factor
            scale_factor = ast.literal_eval(
                om.root.iom_measurements.full_scan_field_of_view.scale_factor
            )

            # Again, the fov is for the larger dimension of the scan
            n_pixels = max(sizes[0], sizes[1])
            value = fov / scale_factor / n_pixels
            scale, unit = _convert_scale_units(value, "m")
            for i in [0, 1]:
                scales[i] = scale
                units[i] = unit

        else:  # pragma: no cover
            _logger.warning("The scale of the navigation axes cannot be read.")

    # Try to set the pixel size for the diffraction patterns.
    # In newer EMPAD versions, the diffraction scale is no longer recorded in the xml. We provide the option to set the scale by providing a q_calibration value.
    if q_calibration is not None:
        dp_pixel_size = q_calibration
        dp_unit = "1/nm"
    # If q_calibration is not provided, check if the calibrated_pixelsize is available in the xml for backward compatibility.
    elif om.has_item("root.iom_measurements.calibrated_pixelsize"):
        dp_pixel_size = float(om.root.iom_measurements.calibrated_pixelsize) * 1e9
        dp_unit = "1/nm"
    else:
        # Fall back to pixel units if the diffraction scale cannot be read from the xml and q_calibration is not provided.
        dp_pixel_size = 1
        dp_unit = None
        _logger.info(
            "The scale of the diffraction axes is not set. The calibration can be set using the `q_calibration` parameter."
        )
    # Write the scale and unit for the axes.
    for i in [-1, -2]:
        scales[i] = dp_pixel_size
        units[i] = dp_unit
    for i in range(len(names)):
        if sizes[i] > 1:
            axes.append(
                {
                    "size": sizes[i],
                    "index_in_array": index_in_array,
                    "name": names[i],
                    "scale": scales[i],
                    "offset": origins[i] * scales[i],
                    "units": units[i],
                    "navigate": navigate[i],
                }
            )
            index_in_array += 1

    data = _read_raw(info, os.path.join(dname, info["raw_filename"]), lazy=lazy)

    # Remove NaN values if requested. Will add a few seconds for the loading time.
    if remove_nans:
        data = np.nan_to_num(data)

    dictionary = {
        "data": data.squeeze(),
        "axes": axes,
        "metadata": md,
        "original_metadata": om.to_dict(),
    }

    return [
        dictionary,
    ]


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np

from rsciio._docstrings import (
    CHUNKS_DOC,
    FILENAME_DOC,
    LAZY_DOC,
    RETURNS_DOC,
    SIGNAL_DOC,
    UNSUPPORTED_METADATA_DOC,
)
from rsciio.utils import file
from rsciio.utils._array import is_dask_array


def inspect_npy_file(filename):
    """
    Inspect a .npy file to extract metadata such as data offset, shape, and dtype.

    Parameters
    ----------
    filename : str
        Path to the .npy file to inspect.

    Returns
    -------
    tuple
        A tuple containing:
        - data_offset (int): The byte offset where the data starts.
        - shape (tuple): The shape of the array stored in the file.
        - dtype (str): The data type of the array elements.

    Example
    -------
    >>> offset, shape, dtype = inspect_npy_file('example.npy')
    >>> print(f"Data offset: {offset} bytes")
    >>> print(f"Shape: {shape}")
    >>> print(f"Dtype: {dtype}")
    """
    with open(filename, "rb") as f:
        return file.inspect_npy_bytes(f)


def file_writer(filename, signal, **kwargs):
    """
    Write data to npy files.

    Parameters
    ----------
    %s
    %s
    **kwargs : dict, optional
        Additional keyword arguments passed to :func:`numpy.save`.

    %s
    """
    array = signal["data"]
    if is_dask_array(array):
        raise TypeError("Lazy signal are not supported for writing to npy files.")

    np.save(filename, array, **kwargs)


file_writer.__doc__ %= (
    FILENAME_DOC.replace("read", "write to"),
    SIGNAL_DOC,
    UNSUPPORTED_METADATA_DOC,
)


def file_reader(filename, lazy=False, chunks="auto", navigation_axes=None, **kwargs):
    """
    Read data from npy files.

    Parameters
    ----------
    %s
    %s
    %s
    navigation_axes : list, optional
        List of axes that should be treated as navigation axes. If not provided,
        all axes will be treated as signal axes.
    **kwargs : dict, optional
        Pass keyword arguments to the :func:`numpy.load`, when
        lazy is False, otherwise to :func:`rsciio.utils.file.memmap_distributed`.

    %s

    %s

    Examples
    --------
    To load a numpy file with lazy loading and specified chunks:

    >>> from rsciio.numpy import file_reader
    >>> d = file_reader('data.npy', lazy=True, chunks=("auto", "auto", 250, 250), navigation_axes=[0, 1])
    >>> d[0]['data']
    dask.array<array, shape=(1000, 1000, 500, 500), dtype=float64, chunksize=(100, 100, 250, 250)>
    """
    if lazy:
        offset, shape, dtype = inspect_npy_file(filename)
        data = file.memmap_distributed(
            filename,
            offset=offset,
            shape=shape,
            dtype=np.dtype(dtype),
            chunks=chunks,
            **kwargs,
        )
    else:
        data = np.load(filename, **kwargs)

    axes = []
    index_in_array = 0
    if navigation_axes is None:
        navigation_axes = []
    for axis, length in enumerate(data.shape):
        if length > 1:
            axes.append(
                {
                    "size": length,
                    "index_in_array": index_in_array,
                    "name": "",
                    "scale": 1.0,
                    "offset": 0.0,
                    "units": "",
                    "navigate": axis in navigation_axes,
                }
            )
            index_in_array += 1

    return [
        {
            "data": data.squeeze(),
            "axes": axes,
            "metadata": {},
            "original_metadata": {},
        },
    ]


file_reader.__doc__ %= (
    FILENAME_DOC,
    LAZY_DOC,
    CHUNKS_DOC,
    RETURNS_DOC,
    UNSUPPORTED_METADATA_DOC,
)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import importlib
import os
import tempfile
from datetime import datetime
from time import perf_counter, sleep

import numpy as np
import numpy.testing as npt
import pytest

from rsciio.utils._tests import assert_deep_almost_equal

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
mrcz = pytest.importorskip("mrcz", reason="mrcz not installed")


# ==============================================================================
# MRCZ Test
#
# Internal python-only test. Build a random image and save and re-load it.
# ==============================================================================
tmpDir = tempfile.gettempdir()

MAX_ASYNC_TIME = 10.0
dtype_list = ["float32", "int8", "int16", "uint16", "complex64"]


def _generate_parameters():
    parameters = []
    for dtype in dtype_list:
        for compressor in [None, "zstd", "lz4"]:
            for clevel in [1, 9]:
                for lazy in [True, False]:
                    parameters.append([dtype, compressor, clevel, lazy])
    return parameters


class TestPythonMrcz:
    def setup_method(self, method):
        pass

    def compareSaveLoad(
        self,
        testShape,
        dtype="int8",
        compressor=None,
        clevel=1,
        lazy=False,
        do_async=False,
        **kwargs,
    ):
        # This is the main function which reads and writes from disk.
        mrcName = os.path.join(tmpDir, "testMage_{}_lazy_{}.mrcz".format(dtype, lazy))

        dtype = np.dtype(dtype)
        if dtype == "float32" or dtype == "float64":
            testData = np.random.normal(size=testShape).astype(dtype)
        elif dtype == "complex64" or dtype == "complex128":
            testData = np.random.normal(size=testShape).astype(
                dtype
            ) + 1.0j * np.random.normal(size=testShape).astype(dtype)
        else:  # integers
            testData = np.random.randint(10, size=testShape).astype(dtype)

        testSignal = hs.signals.Signal2D(testData)
        if lazy:
            testSignal = testSignal.as_lazy()

        # Add "File" metadata to testSignal
        testSignal.metadata.General.add_dictionary(
            {
                "FileIO": {
                    "0": {
                        "operation": "load",
                        "hyperspy_version": hs.__version__,
                        "io_plugin": "rsciio.mrcz",
                        "timestamp": datetime.now().astimezone().isoformat(),
                    }
                }
            }
        )

        # Unfortunately one cannot iterate over axes_manager in a Pythonic way
        # for axis in testSignal.axes_manager:
        testSignal.axes_manager[0].name = "z"
        testSignal.axes_manager[0].scale = np.random.uniform(low=0.0, high=1.0)
        testSignal.axes_manager[0].units = "nm"
        testSignal.axes_manager[1].name = "x"
        testSignal.axes_manager[1].scale = np.random.uniform(low=0.0, high=1.0)
        testSignal.axes_manager[1].units = "nm"
        testSignal.axes_manager[2].name = "y"
        testSignal.axes_manager[2].scale = np.random.uniform(low=0.0, high=1.0)
        testSignal.axes_manager[2].units = "nm"

        # Meta-data that goes into MRC fixed header
        testSignal.metadata.set_item("Acquisition_instrument.TEM.beam_energy", 300.0)
        # Meta-data that goes into JSON extended header
        testSignal.metadata.set_item("Acquisition_instrument.TEM.magnification", 25000)
        testSignal.metadata.set_item(
            "Signal.Noise_properties.Variance_linear_model.gain_factor", 1.0
        )

        testSignal.save(
            mrcName, compressor=compressor, clevel=clevel, do_async=do_async, **kwargs
        )
        if do_async:
            # Poll file on disk since we don't return the
            # concurrent.futures.Future
            t_stop = perf_counter() + MAX_ASYNC_TIME
            sleep(0.005)
            while perf_counter() < t_stop:
                try:
                    fh = open(mrcName, "a")
                    fh.close()
                    break
                except IOError:
                    sleep(0.001)
            print(  # noqa: T201
                "Time to save file: {} s".format(
                    perf_counter() - (t_stop - MAX_ASYNC_TIME)
                )
            )
            sleep(0.1)

        reSignal = hs.load(mrcName)
        try:
            os.remove(mrcName)
        except IOError:
            print("Warning: file {} left on disk".format(mrcName))  # noqa: T201

        # change file timestamp to make the metadata of both signals equal
        testSignal.metadata.General.FileIO.Number_0.timestamp = (
            reSignal.metadata.General.FileIO.Number_0.timestamp
        )

        npt.assert_array_almost_equal(testSignal.data.shape, reSignal.data.shape)
        npt.assert_array_almost_equal(testSignal.data, reSignal.data)
        npt.assert_almost_equal(
            testSignal.metadata.Acquisition_instrument.TEM.beam_energy,
            reSignal.metadata.Acquisition_instrument.TEM.beam_energy,
        )
        npt.assert_almost_equal(
            testSignal.metadata.Acquisition_instrument.TEM.magnification,
            reSignal.metadata.Acquisition_instrument.TEM.magnification,
        )

        for aName in ["x", "y", "z"]:
            npt.assert_equal(
                testSignal.axes_manager[aName].size, reSignal.axes_manager[aName].size
            )
            npt.assert_almost_equal(
                testSignal.axes_manager[aName].scale, reSignal.axes_manager[aName].scale
            )

        if dtype == "complex64":
            assert isinstance(reSignal, hs.signals.ComplexSignal2D)
        else:
            assert isinstance(reSignal, hs.signals.Signal2D)

        # delete last load operation from reSignal metadata so we can compare
        del reSignal.metadata.General.FileIO.Number_2
        assert_deep_almost_equal(
            testSignal.axes_manager.as_dictionary(),
            reSignal.axes_manager.as_dictionary(),
        )
        assert_deep_almost_equal(
            testSignal.metadata.as_dictionary(), reSignal.metadata.as_dictionary()
        )

        return reSignal

    @pytest.mark.parametrize(
        ("dtype", "compressor", "clevel", "lazy"), _generate_parameters()
    )
    def test_MRC(self, dtype, compressor, clevel, lazy):
        blosc = importlib.util.find_spec("blosc")

        if blosc is None and compressor is not None:
            with pytest.raises(ImportError):
                self.compareSaveLoad(
                    [2, 64, 32],
                    dtype=dtype,
                    compressor=compressor,
                    clevel=clevel,
                    lazy=lazy,
                )
        else:
            self.compareSaveLoad(
                [2, 64, 32],
                dtype=dtype,
                compressor=compressor,
                clevel=clevel,
                lazy=lazy,
            )

    @pytest.mark.parametrize("dtype", dtype_list)
    def test_Async(self, dtype):
        _ = pytest.importorskip("blosc", reason="skipping test_async, requires blosc")
        t_start = perf_counter()
        self.compareSaveLoad(
            [2, 64, 32], dtype=dtype, compressor="zstd", clevel=1, do_async=True
        )
        print(  # noqa: T201
            "MRCZ Asynchronous test finished in {} s".format(perf_counter() - t_start)
        )



import pytest

from rsciio.utils._distributed import get_chunk_slice


@pytest.mark.parametrize(
    "shape", ((10, 20, 30, 512, 512), (20, 30, 512, 512), (10, 512, 512), (512, 512))
)
def test_get_chunk_slice(shape):
    chunk_arr, chunk = get_chunk_slice(shape=shape, chunks=-1)  # 1 chunk
    assert chunk_arr.shape == (1,) * len(shape) + (len(shape), 2)
    assert chunk == tuple([(i,) for i in shape])

    chunks = (1,) * (len(shape) - 2) + (-1, -1)
    # Everything is 1 chunk
    chunk_arr, chunk = get_chunk_slice(shape=shape, chunks=chunks)
    assert chunk_arr.shape == shape[:-2] + (1, 1) + (len(shape), 2)
    assert chunk == (
        tuple([(1,) * i for i in shape[:-2]]) + tuple([(i,) for i in shape[-2:]])
    )



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import ast
import datetime
import logging
import warnings

import h5py
import numpy as np
from packaging.version import Version

from rsciio._docstrings import SHOW_PROGRESSBAR_DOC
from rsciio.utils._array import is_dask_array
from rsciio.utils._tools import ensure_unicode

version = "3.3"

default_version = Version(version)

not_valid_format = "The file is not a valid HyperSpy hdf5 file"

_logger = logging.getLogger(__name__)

# Functions to flatten and unflatten the data to allow for storing
# ragged arrays in hdf5 with dimensionality higher than 1


def flatten_data(x, is_hdf5=False):
    new_data = np.empty(shape=x.shape, dtype=object)
    shapes = np.empty(shape=x.shape, dtype=object)
    for i in np.ndindex(x.shape):
        data_ = np.array(x[i]).ravel()
        if np.issubdtype(data_.dtype, np.dtype("U")):
            if is_hdf5:
                # h5py doesn't support numpy unicode dtype, convert to
                # compatible dtype
                new_data[i] = data_.astype(h5py.string_dtype())
            else:
                # Convert to list to save ragged array of array with string dtype
                new_data[i] = data_.tolist()
        else:
            new_data[i] = data_
        shapes[i] = np.array(np.array(x[i]).shape)
    return new_data, shapes


def unflatten_data(data, shape, is_hdf5=False):
    new_data = np.empty(shape=data.shape, dtype=object)
    for i in np.ndindex(new_data.shape):
        try:
            # For hspy file, ragged array of string are saving with
            # "h5py.string_dtype()" type and we need to convert it back
            # to numpy unicode type. The only to know when this needs to be
            # done is look at the numpy metadata
            # This numpy feature is "not well supported in numpy"
            # https://numpy.org/doc/stable/reference/generated/numpy.dtype.metadata.html
            convert_to_unicode = (
                is_hdf5
                and data.dtype is not None
                and data.dtype.metadata.get("vlen") is not None
                and issubclass(data.dtype.metadata["vlen"].metadata.get("vlen"), str)
            )
        except (AttributeError, KeyError):
            # AttributeError in case `dtype.metadata`` is None (most of the time)
            # KeyError in case "vlen" is not a key
            convert_to_unicode = False
        data_ = data[i].astype("U") if convert_to_unicode else data[i]
        new_data[i] = np.reshape(np.array(data_), shape[i])
    return new_data


# ---------------------------------


def get_signal_chunks(shape, dtype, signal_axes=None, target_size=1e6):
    """
    Function that calculates chunks for the signal, preferably at least one
    chunk per signal space.

    Parameters
    ----------
    shape : tuple
        The shape of the dataset to be stored / chunked.
    dtype : {dtype, string}
        The numpy dtype of the data.
    signal_axes : {None, iterable of ints}
        The axes defining "signal space" of the dataset. If None, the default
        h5py chunking is performed.
    target_size : int
        The target number of bytes for one chunk
    """
    typesize = np.dtype(dtype).itemsize
    if shape == (0,) or signal_axes is None:
        # enable autochunking from h5py
        return True

    # largely based on the guess_chunk in h5py
    bytes_per_signal = np.prod([shape[i] for i in signal_axes]) * typesize
    signals_per_chunk = int(np.floor_divide(target_size, bytes_per_signal))
    navigation_axes = tuple(i for i in range(len(shape)) if i not in signal_axes)
    num_nav_axes = len(navigation_axes)
    num_signals = np.prod([shape[i] for i in navigation_axes])
    if signals_per_chunk < 2 or num_nav_axes == 0:
        # signal is larger than chunk max
        chunks = [s if i in signal_axes else 1 for i, s in enumerate(shape)]
        return tuple(chunks)
    elif signals_per_chunk > num_signals:
        return shape
    else:
        # signal is smaller than chunk max
        # Index of axes with size smaller than required to make all chunks equal
        small_idx = []
        # Sizes of axes with size smaller than required to make all chunks equal
        small_sizes = []
        iterate = True
        while iterate:
            iterate = False
            # Calculate the size of the chunks of the axes not in `small_idx`
            # The process is iterative because `nav_axes_chunks` can be bigger
            # than some axes sizes. If that is the case, the value must be
            # recomputed at the next iteration after having added the "offending"
            # axes to `small_idx`
            nav_axes_chunks = int(
                np.floor(
                    (signals_per_chunk / np.prod(small_sizes))
                    ** (1 / (num_nav_axes - len(small_sizes)))
                )
            )
            for index, size in enumerate(shape):
                if (
                    index not in (list(signal_axes) + small_idx)
                    and size < nav_axes_chunks
                ):
                    small_idx.append(index)
                    small_sizes.append(size)
                    iterate = True
        chunks = [
            s if i in signal_axes or i in small_idx else nav_axes_chunks
            for i, s in enumerate(shape)
        ]
        return tuple(int(x) for x in chunks)


class HierarchicalReader:
    """A generic Reader class for reading data from hierarchical file types."""

    _file_type = ""
    _is_hdf5 = False

    def __init__(self, file):
        """
        Initializes a general reader for hierarchical signals.

        Parameters
        ----------
        file: str
            A file to be read.
        """
        self.file = file
        # Getting version also check that this is a hyperspy format
        self.version = self.get_format_version()
        self.Dataset = None
        self.Group = None

        if self.version > Version(version):
            warnings.warn(
                "This file was written using a newer version of the "
                f"HyperSpy {self._file_type} file format. I will attempt to "
                "load it, but, if I fail, it is likely that I will be more "
                "successful at this and other tasks if you upgrade me."
            )

    def get_format_version(self):
        """Return the format version."""
        if "file_format_version" in self.file.attrs:
            version = self.file.attrs["file_format_version"]
            if isinstance(version, bytes):
                version = version.decode()
            if isinstance(version, float):
                version = str(round(version, 2))
        elif "Experiments" in self.file:
            # Chances are that this is a HSpy hdf5 file version 1.0
            version = "1.0"
        elif "Analysis" in self.file:
            # Starting version 2.0 we have "Analysis" field as well
            version = "2.0"
        else:
            raise IOError(not_valid_format)

        return Version(version)

    def read(self, lazy):
        """
        Read all data, metadata, models.

        Parameters
        ----------
        lazy : bool
            Return data as lazy signal.

        Raises
        ------
        IOError
            Raise an IOError when the file can't be read, if the file
            doesn't follow hspy format specification, etc.

        Returns
        -------
        list of dict
            A list of dictionary, which can be used to create a hspy signal.
        """
        models_with_signals = []
        standalone_models = []

        if "Analysis/models" in self.file:
            try:
                m_gr = self.file["Analysis/models"]
                for model_name in m_gr:
                    if "_signal" in m_gr[model_name].attrs:
                        key = m_gr[model_name].attrs["_signal"]
                        # del m_gr[model_name].attrs['_signal']
                        res = self._group2dict(m_gr[model_name], lazy=lazy)
                        del res["_signal"]
                        models_with_signals.append((key, {model_name: res}))
                    else:
                        standalone_models.append(
                            {model_name: self._group2dict(m_gr[model_name], lazy=lazy)}
                        )
            except TypeError:
                raise IOError(not_valid_format)

        experiments = []
        exp_dict_list = []

        if "Experiments" in self.file:
            for ds in self.file["Experiments"]:
                if isinstance(self.file["Experiments"][ds], self.Group):
                    if "data" in self.file["Experiments"][ds]:
                        experiments.append(ds)
            # Parse the file
            for experiment in experiments:
                exg = self.file["Experiments"][experiment]
                exp = self.group2signaldict(exg, lazy)
                # assign correct models, if found:
                _tmp = {}
                for key, _dict in reversed(models_with_signals):
                    if key == exg.name:
                        _tmp.update(_dict)
                        models_with_signals.remove((key, _dict))
                exp["models"] = _tmp

                exp_dict_list.append(exp)

        for _, m in models_with_signals:
            standalone_models.append(m)

        exp_dict_list.extend(standalone_models)

        if not len(exp_dict_list):
            raise IOError(f"This is not a valid {self._file_type} file.")

        return exp_dict_list

    def _read_array(self, group, dataset_key):
        # This is a workaround for the lack of support for n-d ragged array
        # in h5py and zarr. There is work in progress for implementation in zarr:
        # https://github.com/zarr-developers/zarr-specs/issues/62 which may be
        # relevant to implement here when available
        data = group[dataset_key]
        key = f"_ragged_shapes_{dataset_key}"
        if "ragged_shapes" in group:
            # For file saved with rosettaSciIO <= 0.1
            # rename from `ragged_shapes` to `_ragged_shapes_{key}` in v3.3
            key = "ragged_shapes"
        if key in group:
            import dask.array as da

            ragged_shape = group[key]
            # Use same chunks as data so that apply_gufunc doesn't rechunk
            # Reduces the transfer of data between workers which
            # significantly improves performance for distributed loading
            data = da.from_array(data, chunks=data.chunks)
            shapes = da.from_array(ragged_shape, chunks=data.chunks)

            data = da.apply_gufunc(
                unflatten_data,
                "(),()->()",
                data,
                shapes,
                is_hdf5=self._is_hdf5,
                output_dtypes=object,
            )
        return data

    def group2signaldict(self, group, lazy=False):
        """
        Reads a h5py/zarr group and returns a signal dictionary.

        Parameters
        ----------
        group : :py:class:`h5py.Group` or :py:class:`zarr.hierarchy.Group`
            A group following hspy specification.
        lazy : bool, optional
            Return the data as dask array. The default is False.

        Raises
        ------
        IOError
            Raise an IOError when the group can't be read, if the group
            doesn't follow hspy format specification, etc.

        """
        if self.version < Version("1.2"):
            metadata = "mapped_parameters"
            original_metadata = "original_parameters"
        else:
            metadata = "metadata"
            original_metadata = "original_metadata"

        exp = {
            "metadata": self._group2dict(group[metadata], lazy=lazy),
            "original_metadata": self._group2dict(group[original_metadata], lazy=lazy),
        }
        if "attributes" in group:
            # RosettaSciIO version is > 0.1
            exp["attributes"] = self._group2dict(group["attributes"], lazy=lazy)
        else:
            exp["attributes"] = {}
        if "package" in group.attrs:
            # HyperSpy version is >= 1.5
            exp["package"] = group.attrs["package"]
            exp["package_version"] = group.attrs["package_version"]
        else:
            # Prior to v1.4 we didn't store the package information. Since there
            # were already external package we cannot assume any package provider so
            # we leave this empty.
            exp["package"] = ""
            exp["package_version"] = ""

        data = self._read_array(group, "data")
        if lazy:
            if not is_dask_array(data):
                import dask.array as da

                data = da.from_array(data, chunks=data.chunks)
            exp["attributes"]["_lazy"] = True
        else:
            if is_dask_array(data):
                data = data.compute()
            data = np.asanyarray(data)
            exp["attributes"]["_lazy"] = False
        exp["data"] = data
        axes = []
        for i in range(len(exp["data"].shape)):
            try:
                axes.append(self._group2dict(group[f"axis-{i}"]))
                axis = axes[-1]
                for key, item in axis.items():
                    if isinstance(item, np.bool_):
                        axis[key] = bool(item)
                    else:
                        axis[key] = ensure_unicode(item)
            except KeyError:
                break
        if len(axes) != len(exp["data"].shape):  # broke from the previous loop
            try:
                axes = [
                    i
                    for k, i in sorted(
                        iter(
                            self._group2dict(
                                group["_list_" + str(len(exp["data"].shape)) + "_axes"],
                                lazy=lazy,
                            ).items()
                        )
                    )
                ]
            except KeyError:
                raise IOError(not_valid_format)
        exp["axes"] = axes
        if "learning_results" in group.keys():
            exp["attributes"]["learning_results"] = self._group2dict(
                group["learning_results"], lazy=lazy
            )
        if "peak_learning_results" in group.keys():
            exp["attributes"]["peak_learning_results"] = self._group2dict(
                group["peak_learning_results"], lazy=lazy
            )

        # If the title was not defined on writing the Experiment is
        # then called __unnamed__. The next "if" simply sets the title
        # back to the empty string
        if "General" in exp["metadata"] and "title" in exp["metadata"]["General"]:
            if "__unnamed__" == exp["metadata"]["General"]["title"]:
                exp["metadata"]["General"]["title"] = ""

        if self.version < Version("1.1"):
            # Load the decomposition results written with the old name,
            # mva_results
            if "mva_results" in group.keys():
                exp["attributes"]["learning_results"] = self._group2dict(
                    group["mva_results"], lazy=lazy
                )
            if "peak_mva_results" in group.keys():
                exp["attributes"]["peak_learning_results"] = self._group2dict(
                    group["peak_mva_results"], lazy=lazy
                )
            # Replace the old signal and name keys with their current names
            if "signal" in exp["metadata"]:
                if "Signal" not in exp["metadata"]:
                    exp["metadata"]["Signal"] = {}
                exp["metadata"]["Signal"]["signal_type"] = exp["metadata"]["signal"]
                del exp["metadata"]["signal"]

            if "name" in exp["metadata"]:
                if "General" not in exp["metadata"]:
                    exp["metadata"]["General"] = {}
                exp["metadata"]["General"]["title"] = exp["metadata"]["name"]
                del exp["metadata"]["name"]

        if self.version < Version("1.2"):
            if "_internal_parameters" in exp["metadata"]:
                exp["metadata"]["_HyperSpy"] = exp["metadata"]["_internal_parameters"]
                del exp["metadata"]["_internal_parameters"]
                if "stacking_history" in exp["metadata"]["_HyperSpy"]:
                    exp["metadata"]["_HyperSpy"]["Stacking_history"] = exp["metadata"][
                        "_HyperSpy"
                    ]["stacking_history"]
                    del exp["metadata"]["_HyperSpy"]["stacking_history"]
                if "folding" in exp["metadata"]["_HyperSpy"]:
                    exp["metadata"]["_HyperSpy"]["Folding"] = exp["metadata"][
                        "_HyperSpy"
                    ]["folding"]
                    del exp["metadata"]["_HyperSpy"]["folding"]
            if "Variance_estimation" in exp["metadata"]:
                if "Noise_properties" not in exp["metadata"]:
                    exp["metadata"]["Noise_properties"] = {}
                exp["metadata"]["Noise_properties"]["Variance_linear_model"] = exp[
                    "metadata"
                ]["Variance_estimation"]
                del exp["metadata"]["Variance_estimation"]
            if "TEM" in exp["metadata"]:
                if "Acquisition_instrument" not in exp["metadata"]:
                    exp["metadata"]["Acquisition_instrument"] = {}
                exp["metadata"]["Acquisition_instrument"]["TEM"] = exp["metadata"][
                    "TEM"
                ]
                del exp["metadata"]["TEM"]
                tem = exp["metadata"]["Acquisition_instrument"]["TEM"]
                if "EELS" in tem:
                    if "dwell_time" in tem:
                        tem["EELS"]["dwell_time"] = tem["dwell_time"]
                        del tem["dwell_time"]
                    if "dwell_time_units" in tem:
                        tem["EELS"]["dwell_time_units"] = tem["dwell_time_units"]
                        del tem["dwell_time_units"]
                    if "exposure" in tem:
                        tem["EELS"]["exposure"] = tem["exposure"]
                        del tem["exposure"]
                    if "exposure_units" in tem:
                        tem["EELS"]["exposure_units"] = tem["exposure_units"]
                        del tem["exposure_units"]
                    if "Detector" not in tem:
                        tem["Detector"] = {}
                    tem["Detector"] = tem["EELS"]
                    del tem["EELS"]
                if "EDS" in tem:
                    if "Detector" not in tem:
                        tem["Detector"] = {}
                    if "EDS" not in tem["Detector"]:
                        tem["Detector"]["EDS"] = {}
                    tem["Detector"]["EDS"] = tem["EDS"]
                    del tem["EDS"]
                del tem
            if "SEM" in exp["metadata"]:
                if "Acquisition_instrument" not in exp["metadata"]:
                    exp["metadata"]["Acquisition_instrument"] = {}
                exp["metadata"]["Acquisition_instrument"]["SEM"] = exp["metadata"][
                    "SEM"
                ]
                del exp["metadata"]["SEM"]
                sem = exp["metadata"]["Acquisition_instrument"]["SEM"]
                if "EDS" in sem:
                    if "Detector" not in sem:
                        sem["Detector"] = {}
                    if "EDS" not in sem["Detector"]:
                        sem["Detector"]["EDS"] = {}
                    sem["Detector"]["EDS"] = sem["EDS"]
                    del sem["EDS"]
                del sem

            if (
                "Sample" in exp["metadata"]
                and "Xray_lines" in exp["metadata"]["Sample"]
            ):
                exp["metadata"]["Sample"]["xray_lines"] = exp["metadata"]["Sample"][
                    "Xray_lines"
                ]
                del exp["metadata"]["Sample"]["Xray_lines"]

            for key in ["title", "date", "time", "original_filename"]:
                if key in exp["metadata"]:
                    if "General" not in exp["metadata"]:
                        exp["metadata"]["General"] = {}
                    exp["metadata"]["General"][key] = exp["metadata"][key]
                    del exp["metadata"][key]
            for key in ["record_by", "signal_origin", "signal_type"]:
                if key in exp["metadata"]:
                    if "Signal" not in exp["metadata"]:
                        exp["metadata"]["Signal"] = {}
                    exp["metadata"]["Signal"][key] = exp["metadata"][key]
                    del exp["metadata"][key]

        if self.version < Version("3.0"):
            if "Acquisition_instrument" in exp["metadata"]:
                # Move tilt_stage to Stage.tilt_alpha
                # Move exposure time to Detector.Camera.exposure_time
                if "TEM" in exp["metadata"]["Acquisition_instrument"]:
                    tem = exp["metadata"]["Acquisition_instrument"]["TEM"]
                    exposure = None
                    if "tilt_stage" in tem:
                        tem["Stage"] = {"tilt_alpha": tem["tilt_stage"]}
                        del tem["tilt_stage"]
                    if "exposure" in tem:
                        exposure = "exposure"
                    # Digital_micrograph plugin was parsing to 'exposure_time'
                    # instead of 'exposure': need this to be compatible with
                    # previous behaviour
                    if "exposure_time" in tem:
                        exposure = "exposure_time"
                    if exposure is not None:
                        if "Detector" not in tem:
                            tem["Detector"] = {"Camera": {"exposure": tem[exposure]}}
                        tem["Detector"]["Camera"] = {"exposure": tem[exposure]}
                        del tem[exposure]
                # Move tilt_stage to Stage.tilt_alpha
                if "SEM" in exp["metadata"]["Acquisition_instrument"]:
                    sem = exp["metadata"]["Acquisition_instrument"]["SEM"]
                    if "tilt_stage" in sem:
                        sem["Stage"] = {"tilt_alpha": sem["tilt_stage"]}
                        del sem["tilt_stage"]

        return exp

    def _group2dict(self, group, dictionary=None, lazy=False):
        if dictionary is None:
            dictionary = {}
        for key, value in group.attrs.items():
            if isinstance(value, bytes):
                value = value.decode()
            if isinstance(value, (np.bytes_, str)):
                if value == "_None_":
                    value = None
            elif isinstance(value, np.bool_):
                value = bool(value)
            elif isinstance(value, np.ndarray) and value.dtype.char == "S":
                # Convert strings to unicode
                value = value.astype("U")
                if value.dtype.str.endswith("U1"):
                    value = value.tolist()
            # skip signals - these are handled below.
            if key.startswith("_sig_"):
                pass
            elif key.startswith("_list_empty_"):
                dictionary[key[len("_list_empty_") :]] = []
            elif key.startswith("_tuple_empty_"):
                dictionary[key[len("_tuple_empty_") :]] = ()
            elif key.startswith("_bs_"):
                dictionary[key[len("_bs_") :]] = value.tobytes()
            # The following two elif stataments enable reading date and time from
            # v < 2 of HyperSpy's metadata specifications
            elif key.startswith("_datetime_date"):
                date_iso = datetime.date(
                    *ast.literal_eval(value[value.index("(") :])
                ).isoformat()
                dictionary[key.replace("_datetime_", "")] = date_iso
            elif key.startswith("_datetime_time"):
                date_iso = datetime.time(
                    *ast.literal_eval(value[value.index("(") :])
                ).isoformat()
                dictionary[key.replace("_datetime_", "")] = date_iso
            else:
                dictionary[key] = value
        if not isinstance(group, self.Dataset):
            for key in group.keys():
                if key.startswith("_ragged_shapes_"):
                    # array used to parse ragged array, need to skip it
                    # otherwise, it will wrongly read kwargs when reading
                    # variable length markers as they uses ragged arrays
                    pass
                elif key.startswith("_sig_"):
                    dictionary[key] = self.group2signaldict(group[key])
                elif isinstance(group[key], self.Dataset):
                    data_ = self._read_array(group, key)
                    kn = key
                    if key.startswith("_list_"):
                        data_ = self._parse_iterable(data_)
                        data_ = data_.tolist()
                        kn = key[6:]
                    elif key.startswith("_tuple_"):
                        data_ = self._parse_iterable(data_)
                        data_ = tuple(data_.tolist())
                        kn = key[7:]
                    elif data_.dtype.char == "S":
                        data_ = np.array(data_)
                        try:
                            data_ = data_.astype("U")
                        except UnicodeDecodeError:
                            # There are some strings that must stay in binary,
                            # for example dill pickles. This will obviously also
                            # let "wrong" binary string fail somewhere else...
                            pass
                    elif lazy:
                        import dask.array as da

                        data_ = da.from_array(data_, chunks=data_.chunks)
                    else:
                        data_ = np.array(data_)
                    dictionary[kn] = data_
                elif key.startswith("_hspy_AxesManager_"):
                    dictionary[key] = [
                        i
                        for k, i in sorted(
                            iter(self._group2dict(group[key], lazy=lazy).items())
                        )
                    ]
                elif key.startswith("_list_"):
                    dictionary[key[7 + key[6:].find("_") :]] = [
                        i
                        for k, i in sorted(
                            iter(self._group2dict(group[key], lazy=lazy).items())
                        )
                    ]
                elif key.startswith("_tuple_"):
                    dictionary[key[8 + key[7:].find("_") :]] = tuple(
                        [
                            i
                            for k, i in sorted(
                                iter(self._group2dict(group[key], lazy=lazy).items())
                            )
                        ]
                    )
                else:
                    dictionary[key] = {}
                    self._group2dict(group[key], dictionary[key], lazy=lazy)

        return dictionary

    @staticmethod
    def _parse_iterable(data):
        if h5py.check_string_dtype(data.dtype) and hasattr(data, "asstr"):
            # h5py 3.0 and newer
            # https://docs.h5py.org/en/3.0.0/strings.html
            data = data.asstr()[:]
        return np.array(data)


class HierarchicalWriter:
    """
    An object used to simplify and organize the process for writing a
    Hierarchical signal, such as hspy/zspy format.
    """

    target_size = 1e6
    _unicode_kwds = None
    _is_hdf5 = False

    def __init__(self, file, signal, group, **kwds):
        """Initialize a generic file writer for hierarchical data storage types.

        Parameters
        ----------
        file: str
            The file where the signal is to be saved
        signal: BaseSignal
            A BaseSignal to be saved
        group: Group
            A group to where the experimental data will be saved.
        kwds:
            Any additional keywords used for saving the data.
        """
        self.file = file
        self.signal = signal
        self.group = group
        self.Dataset = None
        self.Group = None
        self.kwds = kwds

    @staticmethod
    def _get_object_dset(*args, **kwargs):  # pragma: no cover
        raise NotImplementedError("This method must be implemented by subclasses.")

    @staticmethod
    def _store_data(*arg):  # pragma: no cover
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    def overwrite_dataset(
        cls,
        group,
        data,
        key,
        signal_axes=None,
        chunks=None,
        show_progressbar=True,
        **kwds,
    ):
        """
        Overwrites a dataset into a hierarchical structure following the h5py
        API.

        Parameters
        ----------
        group : :py:class:`zarr.hierarchy.Group` or :py:class:`h5py.Group`
            The group to write the data to.
        data : Array-like
            The data to be written.
        key : str
            The key for the dataset.
        signal_axes : tuple
            The indexes of the signal axes.
        chunks : tuple, None
            The chunks for the dataset. If ``None`` and saving lazy signal,
            the chunks of the dask array will be used otherwise the chunks
            will be determined by the
            :py:func:`~.io_plugins._hierarchical.get_signal_chunks` function.
        %s
        kwds : dict
            Any additional keywords for to be passed to the
            :py:meth:`h5py.Group.require_dataset` or
            :py:meth:`zarr.hierarchy.Group.require_dataset` method.
        """ % SHOW_PROGRESSBAR_DOC
        if chunks is None:
            if is_dask_array(data):
                # For lazy dataset, by default, we use the current dask chunking
                chunks = tuple([c[0] for c in data.chunks])
            else:
                # If signal_axes=None, use automatic h5py chunking, otherwise
                # optimise the chunking to contain at least one signal per chunk
                chunks = get_signal_chunks(
                    data.shape, data.dtype, signal_axes, cls.target_size
                )
        if np.issubdtype(data.dtype, np.dtype("U")):
            # Saving numpy unicode type is not supported in h5py
            data = data.astype(np.dtype("S"))

        if data.dtype != np.dtype("O"):
            got_data = False
            while not got_data:
                try:
                    these_kwds = kwds.copy()
                    these_kwds.update(
                        dict(
                            shape=data.shape,
                            dtype=data.dtype,
                            exact=True,
                            chunks=chunks,
                        )
                    )

                    # If chunks is True, the `chunks` attribute of `dset` below
                    # contains the chunk shape guessed by h5py
                    dset = group.require_dataset(key, **these_kwds)
                    got_data = True
                except TypeError:
                    # if the shape or dtype/etc do not match,
                    # we delete the old one and create new in the next loop run
                    del group[key]

        _logger.info(f"Chunks used for saving: {chunks}")
        if data.dtype == np.dtype("O"):
            if is_dask_array(data):
                import dask.array as da

                new_data, shapes = da.apply_gufunc(
                    flatten_data,
                    "()->(),()",
                    data,
                    is_hdf5=cls._is_hdf5,
                    output_dtypes=[object, object],
                    allow_rechunk=False,
                )
            else:
                new_data, shapes = flatten_data(data, is_hdf5=cls._is_hdf5)

            dset = cls._get_object_dset(group, new_data, key, chunks, **kwds)
            shape_dset = cls._get_object_dset(
                group, shapes, f"_ragged_shapes_{key}", chunks, dtype=int, **kwds
            )

            cls._store_data(
                (new_data, shapes),
                (dset, shape_dset),
                group,
                (key, f"_ragged_shapes_{key}"),
                (chunks, chunks),
                show_progressbar,
            )
        else:
            cls._store_data(data, dset, group, key, chunks, show_progressbar)

    def write(self):
        self.write_signal(self.signal, self.group, **self.kwds)

    def write_signal(
        self,
        signal,
        group,
        write_dataset=True,
        chunks=None,
        show_progressbar=True,
        **kwds,
    ):
        """Writes a signal dict to a hdf5/zarr group"""
        group.attrs.update(signal["package_info"])

        for i, axis_dict in enumerate(signal["axes"]):
            group_name = f"axis-{i}"
            # delete existing group in case the file have been open in 'a' mode
            # and we are saving a different type of axis, to avoid having
            # incompatible axis attributes from previously saved axis.
            if group_name in group.keys():
                del group[group_name]
            coord_group = group.create_group(group_name)
            self.dict2group(axis_dict, coord_group, **kwds)

        mapped_par = group.require_group("metadata")
        metadata_dict = signal["metadata"]

        if write_dataset:
            self.overwrite_dataset(
                group,
                signal["data"],
                "data",
                signal_axes=[
                    idx
                    for idx, axis in enumerate(signal["axes"])
                    if not axis["navigate"]
                ],
                chunks=chunks,
                show_progressbar=show_progressbar,
                **kwds,
            )

        if default_version < Version("1.2"):
            metadata_dict["_internal_parameters"] = metadata_dict.pop("_HyperSpy")

        self.dict2group(metadata_dict, mapped_par, **kwds)
        original_par = group.require_group("original_metadata")
        self.dict2group(signal["original_metadata"], original_par, **kwds)
        learning_results = group.require_group("learning_results")
        self.dict2group(signal["learning_results"], learning_results, **kwds)
        attributes = group.require_group("attributes")
        self.dict2group(signal["attributes"], attributes, **kwds)

        if signal["models"]:
            model_group = self.file.require_group("Analysis/models")
            self.dict2group(signal["models"], model_group, **kwds)
            for model in model_group.values():
                model.attrs["_signal"] = group.name

    def dict2group(self, dictionary, group, **kwds):
        "Recursive writer of dicts and signals"
        for key, value in dictionary.items():
            _logger.debug("Saving item: {}".format(key))
            if isinstance(value, dict):
                self.dict2group(value, group.require_group(key), **kwds)
            elif isinstance(value, (np.ndarray, self.Dataset)) or is_dask_array(value):
                self.overwrite_dataset(group, value, key, **kwds)

            elif value is None:
                group.attrs[key] = "_None_"

            elif isinstance(value, bytes):
                try:
                    # binary string if has any null characters (otherwise not
                    # supported by hdf5)
                    value.index(b"\x00")
                    group.attrs["_bs_" + key] = np.void(value)
                except ValueError:
                    group.attrs[key] = value.decode()

            elif isinstance(value, str):
                group.attrs[key] = value

            elif isinstance(value, list):
                if len(value):
                    self.parse_structure(key, group, value, "_list_", **kwds)
                else:
                    group.attrs["_list_empty_" + key] = "_None_"

            elif isinstance(value, tuple):
                if len(value):
                    self.parse_structure(key, group, value, "_tuple_", **kwds)
                else:
                    group.attrs["_tuple_empty_" + key] = "_None_"

            else:
                try:
                    group.attrs[key] = value
                except Exception:
                    _logger.exception(
                        "The writer could not write the following "
                        f"information in the file: {key} : {value}"
                    )

    def parse_structure(self, key, group, value, _type, **kwds):
        try:
            # Here we check if there are any signals in the container, as
            # casting a long list of signals to a numpy array takes a very long
            # time. So we check if there are any, and save numpy the trouble
            if np.any([isinstance(t, dict) and "_sig_" in t for t in value]):
                tmp = np.array([[0]])
            else:
                tmp = np.array(value)
        except ValueError:
            tmp = np.array([[0]])

        if np.issubdtype(tmp.dtype, object) or tmp.ndim != 1:
            self.dict2group(
                dict(zip([str(i) for i in range(len(value))], value)),
                group.require_group(_type + str(len(value)) + "_" + key),
                **kwds,
            )
        elif np.issubdtype(tmp.dtype, np.dtype("U")):
            if _type + key in group:
                del group[_type + key]
            group.create_dataset(
                _type + key, shape=tmp.shape, **self._unicode_kwds, **kwds
            )
            group[_type + key][:] = tmp[:]
        else:
            if _type + key in group:
                del group[_type + key]
            group.create_dataset(_type + key, data=tmp, **kwds)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


import json
from pathlib import Path

import numpy as np
import pytest

from rsciio.digitalmicrograph._api import (
    DigitalMicrographReader,
    ImageObject,
    file_reader,
)
from rsciio.tests.generate_dm_testing_files import dm3_data_types, dm4_data_types

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")


TEST_DATA_PATH = Path(__file__).parent / "data" / "digitalmicrograph"
DM_1D_PATH = TEST_DATA_PATH / "1D"
DM_2D_PATH = TEST_DATA_PATH / "2D"
DM_3D_PATH = TEST_DATA_PATH / "3D"


class TestImageObject:
    def setup_method(self, method):
        self.imageobject = ImageObject({}, "")

    def _load_file(self, fname):
        with open(fname, "rb") as f:
            dm = DigitalMicrographReader(f)
            dm.parse_file()
            self.imdict = dm.get_image_dictionaries()
        return [ImageObject(imdict, fname) for imdict in self.imdict]

    def test_get_microscope_name(self):
        fname = DM_2D_PATH / "test_diffraction_pattern_tags_removed.dm3"
        images = self._load_file(fname)
        image = images[0]
        # Should return None because the tags are missing
        assert image._get_microscope_name(image.imdict.ImageTags) is None

        fname = DM_2D_PATH / "test_diffraction_pattern.dm3"
        images = self._load_file(fname)
        image = images[0]
        assert image._get_microscope_name(image.imdict.ImageTags) == "FEI Tecnai"

    def test_get_date(self):
        assert self.imageobject._get_date("11/13/2016") == "2016-11-13"

    def test_get_time(self):
        assert self.imageobject._get_time("6:56:37 pm") == "18:56:37"

    def test_parse_string(self):
        assert self.imageobject._parse_string("") is None
        assert self.imageobject._parse_string("string") == "string"

    def test_parse_string_convert_float(self):
        assert self.imageobject._parse_string("5", False) == "5"
        assert self.imageobject._parse_string("5", True) == 5
        assert self.imageobject._parse_string("Imaging", True) is None


def test_missing_tag():
    fname = DM_2D_PATH / "test_diffraction_pattern_tags_removed.dm3"
    s = hs.load(fname)
    md = s.metadata
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.Camera.exposure, 0.2)
    assert md.General.date == "2014-07-09"
    assert md.General.time == "18:56:37"
    assert md.General.title == "test_diffraction_pattern_tags_removed"


def test_read_TEM_metadata():
    fname = TEST_DATA_PATH / ".." / "tiff" / "test_dm_image_um_unit.dm3"
    s = hs.load(fname)
    md = s.metadata
    assert md.Acquisition_instrument.TEM.acquisition_mode == "TEM"
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.Camera.exposure, 0.5)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.magnification, 51.0)
    assert md.Acquisition_instrument.TEM.microscope == "FEI Tecnai"
    assert md.General.date == "2015-07-20"
    assert md.General.original_filename == "test_dm_image_um_unit.dm3"
    assert md.General.title == "test_dm_image_um_unit"
    assert md.General.time == "18:48:25"
    assert md.Signal.quantity == "Intensity"
    assert md.Signal.signal_type == ""


def test_read_Diffraction_metadata():
    fname = DM_2D_PATH / "test_diffraction_pattern.dm3"
    s = hs.load(fname)
    md = s.metadata
    assert md.Acquisition_instrument.TEM.acquisition_mode == "TEM"
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.Camera.exposure, 0.2)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.camera_length, 320.0)
    assert md.Acquisition_instrument.TEM.microscope == "FEI Tecnai"
    assert md.General.date == "2014-07-09"
    assert md.General.original_filename == "test_diffraction_pattern.dm3"
    assert md.General.title == "test_diffraction_pattern"
    assert md.General.time == "18:56:37"
    assert md.Signal.quantity == "Intensity"
    assert md.Signal.signal_type == ""


def test_read_STEM_metadata():
    fname = DM_2D_PATH / "test_STEM_image.dm3"
    s = hs.load(fname)
    md = s.metadata
    assert md.Acquisition_instrument.TEM.acquisition_mode == "STEM"
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.dwell_time, 3.5e-6)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.camera_length, 135.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.magnification, 225000.0)
    assert md.Acquisition_instrument.TEM.microscope == "FEI Titan"
    assert md.General.date == "2016-08-08"
    assert md.General.original_filename == "test_STEM_image.dm3"
    assert md.General.title == "test_STEM_image"
    assert md.General.time == "16:26:37"
    assert md.Signal.quantity == "Intensity"
    assert md.Signal.signal_type == ""


def test_read_EELS_metadata():
    fname = DM_1D_PATH / "test-EELS_spectrum.dm3"
    s = hs.load(fname)
    md = s.metadata
    assert md.Acquisition_instrument.TEM.acquisition_mode == "STEM"
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
    assert md.Acquisition_instrument.TEM.microscope == "FEI Titan"
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.camera_length, 135.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.magnification, 640000.0)
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Stage.tilt_alpha, 24.95, atol=1e-2
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Stage.x, -0.478619, atol=1e-2
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Stage.y, 0.0554612, atol=1e-2
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Stage.z, 0.036348, atol=1e-2
    )
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.convergence_angle, 21.0)
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EELS.collection_angle, 0.0
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EELS.exposure, 0.0035
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EELS.frame_number, 50
    )
    assert md.Acquisition_instrument.TEM.Detector.EELS.spectrometer == "GIF Quantum ER"
    assert md.Acquisition_instrument.TEM.Detector.EELS.aperture_size == 5.0
    assert md.General.date == "2016-08-08"
    assert md.General.original_filename == "test-EELS_spectrum.dm3"
    assert md.General.title == "EELS Acquire"
    assert md.General.time == "19:35:17"
    assert md.Signal.quantity == "Electrons (Counts)"
    assert md.Signal.signal_type == "EELS"
    np.testing.assert_allclose(
        md.Signal.Noise_properties.Variance_linear_model.gain_factor, 0.1285347
    )
    np.testing.assert_allclose(
        md.Signal.Noise_properties.Variance_linear_model.gain_offset, 0.0
    )
    # tag name is "align min. correlation coefficient"
    assert (
        s.original_metadata.ImageList.TagGroup0.ImageTags.EELS.Acquisition.Align_min_correlation_coefficient
        == 0.5
    )


def test_read_SI_metadata():
    fname = DM_3D_PATH / "EELS_SI.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Acquisition_instrument.TEM.acquisition_mode == "STEM"
    assert md.General.date == "2019-05-14"
    assert md.General.time == "20:50:13"
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EELS.aperture_size, 5.0
    )
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.convergence_angle, 21.0)
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EELS.collection_angle, 62.0
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EELS.frame_number, 1
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EELS.dwell_time, 1.9950125e-2
    )


def test_read_EDS_metadata():
    pytest.importorskip("exspy", reason="exspy not installed.")
    fname = DM_1D_PATH / "test-EDS_spectrum.dm3"
    s = hs.load(fname)
    md = s.metadata
    assert md.Acquisition_instrument.TEM.acquisition_mode == "STEM"
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EDS.azimuth_angle, 45.0
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EDS.elevation_angle, 18.0
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa, 130.0
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EDS.live_time, 3.806
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Detector.EDS.real_time, 4.233
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.Stage.tilt_alpha, 24.95, atol=1e-2
    )
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
    assert md.Acquisition_instrument.TEM.microscope == "FEI Titan"
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.camera_length, 135.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.magnification, 320000.0)
    assert md.General.date == "2016-08-08"
    assert md.General.original_filename == "test-EDS_spectrum.dm3"
    assert md.General.title == "EDS Spectrum"
    assert md.General.time == "21:46:19"
    assert md.Signal.quantity == "X-rays (Counts)"
    assert md.Signal.signal_type == "EDS_TEM"
    np.testing.assert_allclose(
        md.Signal.Noise_properties.Variance_linear_model.gain_factor, 1.0
    )
    np.testing.assert_allclose(
        md.Signal.Noise_properties.Variance_linear_model.gain_offset, 0.0
    )
    assert s.axes_manager[-1].units == "keV"


def test_read_MonoCL_pmt_metadata():
    fname = DM_1D_PATH / "test-MonoCL_spectrum-pmt.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Signal.signal_type == "CL"
    assert md.Signal.format == "Spectrum"
    assert md.Signal.quantity == "Intensity (Counts)"
    assert md.General.date == "2020-10-27"
    assert md.General.original_filename == "test-MonoCL_spectrum-pmt.dm4"
    assert md.General.title == "test-CL_spectrum-pmt"
    assert (
        md.Acquisition_instrument.Spectrometer.acquisition_mode == "Serial dispersive"
    )
    assert md.Acquisition_instrument.Detector.detector_type == "PMT"
    assert md.Acquisition_instrument.Spectrometer.Grating.groove_density == 1200
    assert md.Acquisition_instrument.Detector.integration_time == 1.0
    assert md.Acquisition_instrument.Spectrometer.step_size == 0.5
    np.testing.assert_allclose(
        md.Acquisition_instrument.Spectrometer.start_wavelength, 166.233642
    )


def test_read_MonarcCL_pmt_metadata():
    fname = DM_1D_PATH / "test-MonarcCL_spectrum-pmt.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Signal.signal_type == "CL"
    assert md.Signal.format == "Spectrum"
    assert md.Signal.quantity == "Intensity (Counts)"
    assert md.General.date == "2022-01-17"
    assert md.General.original_filename == "test-MonarcCL_spectrum-pmt.dm4"
    assert md.General.title == "CL_15kX_3-60_pmt2s"
    assert (
        md.Acquisition_instrument.Spectrometer.acquisition_mode == "Serial dispersive"
    )
    assert md.Acquisition_instrument.Detector.detector_type == "PMT"
    assert md.Acquisition_instrument.Spectrometer.Grating.groove_density == 300
    assert md.Acquisition_instrument.Detector.integration_time == 2.0
    assert md.Acquisition_instrument.Spectrometer.step_size == 1.0
    np.testing.assert_allclose(
        md.Acquisition_instrument.Spectrometer.start_wavelength, 160.067505
    )


def test_read_MonoCL_ccd_metadata():
    fname = DM_1D_PATH / "test-MonoCL_spectrum-ccd.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Signal.signal_type == "CL"
    assert md.Signal.format == "Spectrum"
    assert md.Signal.quantity == "Intensity (Counts)"
    assert md.General.date == "2020-09-11"
    assert md.General.time == "17:04:19"
    assert md.General.original_filename == "test-MonoCL_spectrum-ccd.dm4"
    assert md.General.title == "test-CL_spectrum-ccd"
    assert md.Acquisition_instrument.Detector.detector_type == "CCD"
    assert md.Acquisition_instrument.SEM.acquisition_mode == "SEM"
    assert md.Acquisition_instrument.SEM.microscope == "Ultra55"
    assert md.Acquisition_instrument.SEM.beam_energy == 5.0
    assert md.Acquisition_instrument.SEM.magnification == 10104.515625
    assert (
        md.Acquisition_instrument.Spectrometer.acquisition_mode == "Parallel dispersive"
    )
    assert md.Acquisition_instrument.Spectrometer.Grating.groove_density == 300.0
    assert md.Acquisition_instrument.Detector.exposure_per_frame == 30.0
    assert md.Acquisition_instrument.Detector.frames == 1.0
    assert md.Acquisition_instrument.Detector.integration_time == 30.0
    np.testing.assert_allclose(
        md.Acquisition_instrument.Spectrometer.central_wavelength, 949.974182
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.saturation_fraction, 0.01861909
    )
    assert md.Acquisition_instrument.Detector.binning == (1, 100)
    assert md.Acquisition_instrument.Detector.processing == "Dark Subtracted"
    assert md.Acquisition_instrument.Detector.sensor_roi == (0, 0, 100, 1336)
    assert md.Acquisition_instrument.Detector.pixel_size == 20.0


def test_read_MonarcCL_ccd_metadata():
    fname = DM_1D_PATH / "test-MonarcCL_spectrum-ccd.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Signal.signal_type == "CL"
    assert md.Signal.format == "Spectrum"
    assert md.Signal.quantity == "Intensity (Counts)"
    assert md.General.date == "2022-01-17"
    assert md.General.time == "16:09:21"
    assert md.General.original_filename == "test-MonarcCL_spectrum-ccd.dm4"
    assert md.General.title == "CL_15kX_3-60_CCD300s_bin2"
    assert md.Acquisition_instrument.Detector.detector_type == "CCD"
    assert md.Acquisition_instrument.SEM.acquisition_mode == "SEM"
    assert md.Acquisition_instrument.SEM.microscope == "Ultra55"
    assert md.Acquisition_instrument.SEM.beam_energy == 3.0
    assert md.Acquisition_instrument.SEM.magnification == 15000.0
    assert (
        md.Acquisition_instrument.Spectrometer.acquisition_mode == "Parallel dispersive"
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.Spectrometer.central_wavelength, 320.049683
    )
    assert md.Acquisition_instrument.Spectrometer.Grating.groove_density == 300.0
    assert md.Acquisition_instrument.Detector.exposure_per_frame == 300.0
    assert md.Acquisition_instrument.Detector.frames == 1.0
    assert md.Acquisition_instrument.Detector.integration_time == 300.0
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.saturation_fraction, 0.08890307
    )
    assert md.Acquisition_instrument.Detector.binning == (2, 100)
    assert md.Acquisition_instrument.Detector.processing == "Dark Subtracted"
    assert md.Acquisition_instrument.Detector.sensor_roi == (0, 0, 100, 1336)
    assert md.Acquisition_instrument.Detector.pixel_size == 20.0


def test_read_MonoCL_SI_metadata():
    fname = DM_2D_PATH / "test-MonoCL_spectrum-SI.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Signal.signal_type == "CL"
    assert md.Signal.format == "Spectrum image"
    assert md.Signal.quantity == "Intensity (Counts)"
    assert md.General.date == "2020-04-11"
    assert md.General.time == "14:41:01"
    assert md.General.original_filename == "test-MonoCL_spectrum-SI.dm4"
    assert md.General.title == "test-CL_spectrum-SI"
    assert md.Acquisition_instrument.Detector.detector_type == "CCD"
    assert md.Acquisition_instrument.SEM.acquisition_mode == "SEM"
    assert md.Acquisition_instrument.SEM.microscope == "Ultra55"
    assert md.Acquisition_instrument.SEM.beam_energy == 5.0
    np.testing.assert_allclose(
        md.Acquisition_instrument.SEM.magnification, 31661.427734
    )
    assert (
        md.Acquisition_instrument.Spectrometer.acquisition_mode == "Parallel dispersive"
    )
    assert md.Acquisition_instrument.Spectrometer.Grating.groove_density == 600.0
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.exposure_per_frame, 0.05
    )
    assert md.Acquisition_instrument.Detector.frames == 1
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.integration_time, 0.05
    )
    assert md.Acquisition_instrument.Detector.pixel_size == 20.0
    np.testing.assert_allclose(
        md.Acquisition_instrument.Spectrometer.central_wavelength, 869.983825
    )
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.saturation_fraction[0], 0.09676377
    )
    assert md.Acquisition_instrument.Detector.binning == (1, 100)
    assert md.Acquisition_instrument.Detector.processing == "Dark Subtracted"
    assert md.Acquisition_instrument.Detector.sensor_roi == (0, 0, 100, 1336)
    assert md.Acquisition_instrument.Spectrum_image.drift_correction_periodicity == 1
    assert (
        md.Acquisition_instrument.Spectrum_image.drift_correction_units == "second(s)"
    )
    assert md.Acquisition_instrument.Spectrum_image.mode == "LineScan"


def test_read_MonarcCL_SI_metadata():
    fname = DM_2D_PATH / "test-MonarcCL_spectrum-SI.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Signal.signal_type == "CL"
    assert md.Signal.format == "Spectrum image"
    assert md.Signal.quantity == "Intensity (Counts)"
    assert md.General.date == "2021-09-16"
    assert md.General.time == "12:06:16"
    assert md.General.original_filename == "test-MonarcCL_spectrum-SI.dm4"
    assert md.General.title == "Monarc_SI_9pix"
    assert md.Acquisition_instrument.Detector.detector_type == "CCD"
    assert md.Acquisition_instrument.SEM.acquisition_mode == "SEM"
    assert md.Acquisition_instrument.SEM.microscope == "Zeiss SEM COM"
    assert md.Acquisition_instrument.SEM.beam_energy == 5.0
    np.testing.assert_allclose(md.Acquisition_instrument.SEM.magnification, 5884.540039)
    assert (
        md.Acquisition_instrument.Spectrometer.acquisition_mode == "Parallel dispersive"
    )
    assert md.Acquisition_instrument.Spectrometer.Grating.groove_density == 1200.0
    assert md.Acquisition_instrument.Spectrometer.entrance_slit_width == 0.256
    assert md.Acquisition_instrument.Spectrometer.bandpass == 0.9984
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.exposure_per_frame, 0.05
    )
    assert md.Acquisition_instrument.Detector.frames == 1
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.integration_time, 0.05
    )
    assert md.Acquisition_instrument.Detector.pixel_size == 20.0
    np.testing.assert_allclose(
        md.Acquisition_instrument.Detector.saturation_fraction[0], 0.004867628
    )
    assert md.Acquisition_instrument.Detector.binning == (2, 400)
    assert md.Acquisition_instrument.Detector.processing == "Dark Subtracted"
    assert md.Acquisition_instrument.Detector.sensor_roi == (0, 0, 400, 1340)
    assert md.Acquisition_instrument.Spectrum_image.mode == "2D Array"


def test_read_MonarcCL_image_metadata():
    fname = DM_2D_PATH / "test-MonarcCL_mono-image.dm4"
    s = hs.load(fname)
    md = s.metadata
    assert md.Signal.signal_type == "CL"
    assert md.Signal.quantity == "Intensity (counts)"
    assert md.General.date == "2021-05-14"
    assert md.General.time == "11:41:07"
    assert md.General.original_filename == "test-MonarcCL_mono-image.dm4"
    assert md.General.title == "MonoCL-image-rebin"
    assert md.Acquisition_instrument.SEM.acquisition_mode == "SEM"
    assert md.Acquisition_instrument.SEM.microscope == "Zeiss SEM COM"
    assert md.Acquisition_instrument.SEM.beam_energy == 3.0
    assert md.Acquisition_instrument.SEM.magnification == 2500.0
    assert md.Acquisition_instrument.SEM.dwell_time == 1e-5
    assert md.Acquisition_instrument.Spectrometer.Grating.groove_density == 300.0
    assert md.Acquisition_instrument.Spectrometer.entrance_slit_width == 0.961
    np.testing.assert_allclose(
        md.Acquisition_instrument.Spectrometer.bandpass, 14.9915999
    )
    assert md.Acquisition_instrument.Detector.pmt_voltage == 1000


def test_location():
    fname_list = [
        "Fei HAADF-DE_location.dm3",
        "Fei HAADF-FR_location.dm3",
        "Fei HAADF-MX_location.dm3",
        "Fei HAADF-UK_location.dm3",
    ]
    s = hs.load(TEST_DATA_PATH / fname_list[0])
    assert s.metadata.General.date == "2016-08-27"
    assert s.metadata.General.time == "20:54:33"
    s = hs.load(TEST_DATA_PATH / fname_list[1])
    assert s.metadata.General.date == "2016-08-27"
    assert s.metadata.General.time == "20:55:20"
    s = hs.load(TEST_DATA_PATH / fname_list[2])
    assert s.metadata.General.date == "2016-08-27"
    assert s.metadata.General.time == "20:55:59"
    s = hs.load(TEST_DATA_PATH / fname_list[3])
    assert s.metadata.General.date == "2016-08-27"
    assert s.metadata.General.time == "20:52:30"


def test_multi_signal():
    fname = DM_2D_PATH / "multi_signal.dm3"
    s = hs.load(fname)

    # Make sure file is read as a list, and exactly two signals are found
    assert isinstance(s, list)
    assert len(s) == 2

    s1, s2 = s

    # First signal is an image, second is a plot
    assert isinstance(s1, hs.signals.Signal2D)
    assert isinstance(s2, hs.signals.Signal1D)

    s1_md_truth = {
        "_HyperSpy": {
            "Folding": {
                "unfolded": False,
                "signal_unfolded": False,
                "original_shape": None,
                "original_axes_manager": None,
            }
        },
        "General": {
            "title": "HAADF",
            "original_filename": "multi_signal.dm3",
            "date": "2019-12-10",
            "time": "15:32:41",
            "authors": "JohnDoe",
            "FileIO": {
                "0": {
                    "operation": "load",
                    "hyperspy_version": hs.__version__,
                    "io_plugin": "rsciio.digitalmicrograph",
                }
            },
        },
        "Signal": {
            "signal_type": "",
            "quantity": "Intensity",
            "Noise_properties": {
                "Variance_linear_model": {"gain_factor": 1.0, "gain_offset": 0.0}
            },
        },
        "Acquisition_instrument": {
            "TEM": {
                "beam_energy": 300.0,
                "Stage": {
                    "tilt_alpha": 0.001951998453075299,
                    "x": 0.07872150000000001,
                    "y": 0.100896,
                    "z": -0.0895279,
                },
                "acquisition_mode": "STEM",
                "beam_current": 0.0,
                "camera_length": 77.0,
                "magnification": 10000000.0,
                "microscope": "Example Microscope",
                "dwell_time": 3.2400001525878905e-05,
            }
        },
        "Sample": {"description": "PrecipitateA"},
    }

    s2_md_truth = {
        "_HyperSpy": {
            "Folding": {
                "unfolded": False,
                "signal_unfolded": False,
                "original_shape": None,
                "original_axes_manager": None,
            }
        },
        "General": {
            "title": "Plot",
            "original_filename": "multi_signal.dm3",
            "FileIO": {
                "0": {
                    "operation": "load",
                    "hyperspy_version": hs.__version__,
                    "io_plugin": "rsciio.digitalmicrograph",
                }
            },
        },
        "Signal": {
            "signal_type": "",
            "quantity": "Intensity",
            "Noise_properties": {
                "Variance_linear_model": {"gain_factor": 1.0, "gain_offset": 0.0}
            },
        },
    }
    # remove timestamps from metadata since these are runtime dependent
    del s1.metadata.General.FileIO.Number_0.timestamp
    del s2.metadata.General.FileIO.Number_0.timestamp

    # make sure the metadata dictionaries are as we expect
    assert s1.metadata.as_dictionary() == s1_md_truth
    assert s2.metadata.as_dictionary() == s2_md_truth

    # rather than testing all of original metadata (huge), use length as a proxy
    assert len(json.dumps(s1.original_metadata.as_dictionary())) == 17779
    assert len(json.dumps(s2.original_metadata.as_dictionary())) == 15024

    # test axes
    assert s1.axes_manager[-1].is_binned is False
    assert s2.axes_manager[-1].is_binned is False

    # simple tests on the data itself:
    assert s1.data.sum() == 949490255
    assert s1.data.shape == (512, 512)
    assert s2.data.sum() == pytest.approx(28.085794, 0.01)
    assert s2.data.shape == (512,)


def generate_parameters():
    parameters = []
    for dim in range(1, 4):
        for key in dm3_data_types.keys():
            subfolder = f"{dim}D"
            filename = TEST_DATA_PATH / subfolder / f"test-{key}.dm3"
            parameters.append(
                {
                    "filename": filename,
                    "subfolder": subfolder,
                    "key": key,
                    "dim": dim,
                }
            )
        for key in dm4_data_types.keys():
            subfolder = f"{dim}D"
            filename = TEST_DATA_PATH / subfolder / f"test-{key}.dm4"
            parameters.append(
                {
                    "filename": filename,
                    "subfolder": subfolder,
                    "key": key,
                    "dim": dim,
                }
            )
    return parameters


## modify this test for axes, maybe extra tests for metadata reads?
@pytest.mark.parametrize("pdict", generate_parameters())
@pytest.mark.parametrize("lazy", (True, False))
def test_data_and_axes(pdict, lazy):
    if lazy:
        if pdict["key"] in (8, 23):  # RGBA
            with pytest.raises(ValueError):
                # RGBA data cannot be lazy loaded
                hs.load(pdict["filename"], lazy=lazy)
            pytest.skip("RGBA data cannot be lazy loaded")
    s = hs.load(pdict["filename"], lazy=lazy)
    if lazy:
        s.compute(close_file=True)
    key = pdict["key"]
    assert s.data.dtype == np.dtype(dm4_data_types[key])
    subfolder = pdict["subfolder"]
    print(pdict["subfolder"])  # noqa: T201
    if subfolder == "1D":
        dat = np.arange(1, 3)
        assert s.axes_manager.signal_shape == (2,)
        assert s.axes_manager.navigation_shape == ()
    elif subfolder == "2D":
        dat = np.arange(1, 5).reshape(2, 2)
        assert s.axes_manager.signal_shape == (2, 2)
        assert s.axes_manager.navigation_shape == ()
    elif subfolder == "3D":
        dat = np.arange(1, 9).reshape(2, 2, 2)
        assert s.axes_manager.signal_shape == (2, 2)
        assert s.axes_manager.navigation_shape == (2,)
    else:
        raise ValueError
    dat = dat.astype(dm4_data_types[key])
    if key in (8, 23):  # RGBA
        dat["A"][:] = 0
    np.testing.assert_array_equal(
        s.data,
        dat,
        err_msg=f"content {subfolder} type {key}: "
        f"\n{str(s.data)} not equal to \n{str(dat)}",
    )


def test_compare_complex_data_lazily():
    # Check that complex data is read correctly when lazy loaded
    # unpacking is the same as for non-lazy data
    s = hs.load(DM_2D_PATH / "test_fft_packed_complex8.dm4", lazy=False)
    s_lazy = hs.load(DM_2D_PATH / "test_fft_packed_complex8.dm4", lazy=True)
    s_lazy.compute(close_file=True)

    assert s.data.dtype == np.complex64
    assert s_lazy.data.dtype == np.complex64
    np.testing.assert_allclose(s.data, s_lazy.data)


def test_axes_bug_for_image():
    fname = DM_2D_PATH / "test_STEM_image.dm3"
    s = hs.load(fname)
    assert s.axes_manager[1].name == "y"


def test_load_stackbuilder_imagestack():
    image_stack = hs.load(TEST_DATA_PATH / "test_stackbuilder_imagestack.dm3")
    data_dimensions = image_stack.data.ndim
    am = image_stack.axes_manager
    axes_dimensions = am.signal_dimension + am.navigation_dimension
    assert data_dimensions == axes_dimensions
    md = image_stack.metadata
    assert md.Acquisition_instrument.TEM.acquisition_mode == "STEM"
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_current, 0.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.beam_energy, 200.0)
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.camera_length, 15.0)
    np.testing.assert_allclose(
        md.Acquisition_instrument.TEM.dwell_time, 0.03000005078125
    )
    np.testing.assert_allclose(md.Acquisition_instrument.TEM.magnification, 200000.0)
    assert md.Acquisition_instrument.TEM.microscope == "JEM-ARM200F"
    assert md.General.date == "2015-05-17"
    assert md.General.original_filename == "test_stackbuilder_imagestack.dm3"
    assert md.General.title == "stackbuilder_test4_16x2"
    assert md.General.time == "17:00:16"
    assert md.Sample.description == "DWNC"
    assert md.Signal.quantity == "Electrons (Counts)"
    assert md.Signal.signal_type == ""
    assert am.signal_axes[0].is_binned is False
    np.testing.assert_allclose(
        md.Signal.Noise_properties.Variance_linear_model.gain_factor, 0.15674974
    )
    np.testing.assert_allclose(
        md.Signal.Noise_properties.Variance_linear_model.gain_offset, 2228741.5
    )


def test_load_packed_complex():
    # Packed complex is typically used for FFT data
    fname = DM_2D_PATH / "test_fft_packed_complex8.dm4"
    file_content = file_reader(fname)
    data_dtype = file_content[0]["data"].dtype
    assert data_dtype == np.complex64



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import gc
import importlib
from pathlib import Path

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")

testfile_dir = (Path(__file__).parent / "data" / "hamamatsu").resolve()

testfile_focus_mode_path = (testfile_dir / "focus_mode.img").resolve()
testfile_operate_mode_path = (testfile_dir / "operate_mode.img").resolve()
testfile_photon_count_path = (testfile_dir / "photon_counting.img").resolve()
testfile_shading_path = (testfile_dir / "shading_file.img").resolve()
testfile_xaxisother_path = (testfile_dir / "xaxis_other.img").resolve()


class TestOperate:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(testfile_operate_mode_path, reader="Hamamatsu")

    @classmethod
    def teardown_class(cls):
        del cls.s
        gc.collect()

    def test_data_deflection(self):
        expected_data_start = [
            [0, 0, 715],
            [246, 161, 475],
            [106, 899, 0],
        ]
        np.testing.assert_allclose(expected_data_start, self.s.isig[:3, :3].data)

    def test_axes_deflection(self):
        axes = self.s.axes_manager
        assert axes.signal_dimension == 2
        assert axes.navigation_dimension == 0
        ax0 = axes[0]
        ax1 = axes[1]
        assert ax0.name == "Wavelength"
        assert ax1.name == "Time"
        assert ax0.size == 672
        assert ax1.size == 512
        assert ax0.units == "nm"
        assert ax1.units == "µs"

        expected_data_start_X = [472.252, 472.33337, 472.41473, 472.4961, 472.57745]
        expected_data_start_Y = [0.0, 0.031080816, 0.062164314, 0.09325048, 0.12433932]
        np.testing.assert_allclose(ax0.axis[:5], expected_data_start_X)
        np.testing.assert_allclose(ax1.axis[:5], expected_data_start_Y)

    def test_original_metadata_no_comment(self):
        original_metadata_no_comment = {
            k: v
            for (k, v) in self.s.original_metadata.as_dictionary().items()
            if k != "Comment"
        }
        expected_metadata = {
            "character_im": "IM",
            "offset_x": 0,
            "offset_y": 0,
            "file_type": "bit32",
            "num_images_in_channel": 0,
            "num_additional_channels": 0,
            "channel_number": 0,
            "timestamp": 0.0,
            "marker": "",
            "additional_info": "",
            "image_width_px": 672,
            "image_height_lines": 512,
            "comment_length": 3311,
        }
        assert expected_metadata == original_metadata_no_comment

    def test_original_metadata_comment(self):
        original_metadata = self.s.original_metadata.Comment.as_dictionary()
        expected_metadata = {
            "Application": {
                "Enconding": "UTF-8",
                "Date": "20/01/2023",
                "Time": "11:34:55.147",
                "Software": "HPD-TA",
                "Application": "2",
                "ApplicationTitle": "High Performance Digital Temporal Analyzer",
                "SoftwareVersion": "9.5 pf10",
                "SoftwareDate": "15.12.2020",
            },
            "Camera": {
                "CameraName": "Orca R2",
                "SerialNumber": "S/N: 891892",
                "Type": "37",
                "ScanMode": "1",
                "Subarray": "0",
                "Binning": "2",
                "BitsPerChannel": "0",
                "HighDynamicRangeMode": "0",
                "ScanSpeed": "0",
                "SubarrayHpos": "0",
                "SubarrayHsize": "0",
                "SubarrayVpos": "0",
                "SubarrayVsize": "0",
                "TriggerMode": "3",
                "TriggerModeKeyVal": "Internal",
                "TriggerPolarity": "1",
                "TriggerPolarityKeyVal": "neg.",
                "CoolerSwitch": "-1",
                "Gain": "0",
                "Prop_SensorMode": "1",
                "Prop_Colortype": "1",
                "Prop_TriggerTimes": "1",
                "Prop_ExposureTimeControl": "2",
                "Prop_TimingMinTriggerInterval": "0.500139",
                "Prop_TimingExposure": "2",
                "Prop_ImageTopOffsetBytes": "0",
                "Prop_ImagePixelType": "2",
                "Prop_BufferRowbytes": "1344",
                "Prop_BufferFramebytes": "688128",
                "Prop_BufferTopOffsetBytes": "0",
                "Prop_BufferPixelType": "2",
                "Prop_RecordFixedBytesPerFile": "256",
                "Prop_RecordFixedBytesPerSession": "784",
                "Prop_RecordFixedBytesPerFrame": "688176",
                "Prop_NumberOfOutputTriggerConnector": "1",
                "Prop_OutputTriggerPolarity": "1",
                "Prop_OutputTriggerActive": "1",
                "Prop_OutputTriggerDelay": "0",
                "Prop_OutputTriggerPeriod": "0.0001",
                "Prop_SystemAlive": "2",
                "Prop_ImageDetectorPixelWidth": "6.45",
                "Prop_ImageDetectorPixelHeight": "6.45",
                "Prop_TimeStampProducer": "2",
                "Prop_FrameStampProducer": "2",
            },
            "Acquisition": {
                "NrExposure": "60",
                "NrTrigger": "0",
                "ExposureTime": "5 s",
                "AcqMode": "4",
                "DataType": "8",
                "DataTypeOfSingleImage": "7",
                "CurveCorr": "0",
                "DefectCorrection": "0",
                "areSource": "0,0,672,512",
                "areGRBScan": "0,0,672,512",
                "pntOrigCh": "0,0",
                "pntOrigFB": "0,0",
                "pntBinning": "2,2",
                "BytesPerPixel": "4",
                "IsLineData": "0",
                "BacksubCorr": "-1",
                "ShadingCorr": "0",
                "ZAxisLabel": "Intensity",
                "ZAxisUnit": "Count",
                "miMirrorRotate": "0",
            },
            "Grabber": {"Type": "5", "SubType": "0"},
            "DisplayLUT": {
                "EntrySize": "9",
                "LowerValue": "764",
                "UpperValue": "3562",
                "LowerValueEx": "0",
                "UpperValueEx": "32767",
                "BitRange": "16x bit",
                "Color": "2",
                "LUTType": "0",
                "LUTInverted": "0",
                "AutoLutInLiveMode": "0",
                "DisplayNegative": "0",
                "Gamma": "1",
                "First812OvlCol": "1",
                "Lut16xShift": "7",
                "Lut16xOvlVal": "3932100",
            },
            "ExternalDevices": {
                "TriggerDelay": "150",
                "PostTriggerTime": "10",
                "ExposureTime": "10",
                "TDStatusCableConnected": "0",
                "ConnectMonitorOut": "0",
                "ConnectResetIn": "0",
                "TriggerMethod": "2",
                "UseDTBE": "0",
                "ExpTimeAddMultiple": "-1",
                "DontSendReset": "0",
                "MultipleOfSweep": "1",
                "A6538Connected": "0",
                "CounterBoardInstalled": "0",
                "MotorizedSlitInstalled": "0",
                "UseSpecAsMono": "0",
                "GPIBInstalled": "-1",
                "CounterBoardIOBase": "0",
                "MotorizedSlitPortID": "1",
                "GPIBIOBase": "0",
            },
            "Streak camera": {
                "UseDevice": "-1",
                "DeviceName": "C5680",
                "PluginName": "M5677",
                "GPIBCableConnected": "-1",
                "GPIBBase": "8",
                "Time Range": "20 us",
                "Mode": "Operate",
                "Gate Mode": "Normal",
                "MCP Gain": "50",
                "Shutter": "Open",
                "Gate Time": "0",
                "Trig. Mode": "Cont",
                "Trigger status": "Ready",
                "Trig. level": "1",
                "Trig. slope": "Rising",
                "FocusTimeOver": "11",
            },
            "Spectrograph": {
                "UseDevice": "-1",
                "DeviceName": "Andor SG",
                "PluginName": "Kymera 328i",
                "GPIBCableConnected": "0",
                "GPIBBase": "13",
                "Wavelength": "500",
                "Grating": "300 g/mm",
                "Blaze": "500",
                "Ruling": "300",
                "Exit Mirror": "Front",
                "Side Ent. Slitw.": "10",
                "Turret": "1",
                "Focus Mirror": "234",
                "Side Entry Iris": "26",
            },
            "Delay box": {
                "UseDevice": "-1",
                "DeviceName": "C4792-01",
                "PluginName": "",
                "GPIBCableConnected": "0",
                "Trig. Mode": "Int. Trig.",
                "Repetition Rate": "1 Hz",
                "L-Pulsewidth": "300 ns",
                "Delay Int.Trig": "-999999999900",
                "Delay Ext.Trig": "0",
                "Dly Mode-Lock": "0",
                "Dly1 DmpMode": "0",
                "Dly2 DmpMode": "0",
            },
            "Delay2 box": {"UseDevice": "0"},
            "Filter wheel": {"UseDevice": "0"},
            "Scaling": {
                "ScalingXType": "2",
                "ScalingXScale": "1",
                "ScalingXUnit": "nm",
                "ScalingXScalingFile": "#1379631,0672",
                "ScalingYType": "2",
                "ScalingYScale": "1",
                "ScalingYUnit": "us",
                "ScalingYScalingFile": "#1382319,0512",
            },
            "Comment": {"UserComment": ""},
        }

        assert expected_metadata == original_metadata

    def test_metadata(self):
        metadata = self.s.metadata

        detector = self.s.metadata.Acquisition_instrument.Detector
        spectrometer = self.s.metadata.Acquisition_instrument.Spectrometer

        assert metadata.General.date == "2023-01-20"
        assert metadata.General.time == "11:34:55"
        assert metadata.General.original_filename == "operate_mode.img"
        assert metadata.General.title == metadata.General.original_filename[:-4]

        assert metadata.Signal.quantity == "Intensity (Counts)"
        if importlib.util.find_spec("lumispy") is None:
            signal_type = ""
        else:  # pragma: no cover
            signal_type = "TransientSpectrum"

        assert metadata.Signal.signal_type == signal_type

        assert isinstance(detector.binning, tuple)
        assert len(detector.binning) == 2
        assert detector.binning[0] == 2
        assert detector.binning[1] == 2
        assert detector.detector_type == "StreakCamera"
        assert detector.model == "C5680"
        assert detector.frames == 60
        np.testing.assert_allclose(detector.integration_time, 300)
        assert detector.processing.background_correction is True
        assert detector.processing.curvature_correction is False
        assert detector.processing.defect_correction is False
        assert detector.processing.shading_correction is False
        np.testing.assert_allclose(detector.time_range, 20)
        assert detector.time_range_units == "µs"
        np.testing.assert_allclose(detector.mcp_gain, 50)
        assert detector.acquisition_mode == "analog_integration"

        np.testing.assert_allclose(spectrometer.entrance_slit_width, 10)
        assert spectrometer.model == "Andor SG"
        assert spectrometer.Grating.blazing_wavelength == 500
        assert spectrometer.Grating.groove_density == 300
        np.testing.assert_allclose(spectrometer.central_wavelength, 500)

    def test_lumispy(self):
        pytest.importorskip("lumispy", reason="lumispy not installed")
        metadata = self.s.metadata
        assert metadata.Signal.signal_type == "TransientSpectrum"


class TestFocus:
    @classmethod
    def setup_class(cls):
        cls.s_focus = hs.load(testfile_focus_mode_path, reader="Hamamatsu")

    @classmethod
    def teardown_class(cls):
        del cls.s_focus
        gc.collect()

    def test_data_focus(self):
        expected_data_end = [[0, 0, 36], [0, 0, 0], [21, 0, 0]]
        np.testing.assert_allclose(expected_data_end, self.s_focus.isig[-3:, -3:].data)

    def test_axes_focus(self):
        axes = self.s_focus.axes_manager
        assert axes.signal_dimension == 2
        assert axes.navigation_dimension == 0
        ax0 = axes[0]
        ax1 = axes[1]
        assert ax0.name == "Wavelength"
        assert ax1.name == "Vertical CCD Position"
        assert ax1.units == "px"
        assert ax0.units == "nm"
        assert ax0.size == 672
        assert ax1.size == 512

        np.testing.assert_allclose(ax1.scale, 1)
        np.testing.assert_allclose(ax1.offset, 0)

        expected_data_start_X = [472.252, 472.33337, 472.41473, 472.4961, 472.57745]
        np.testing.assert_allclose(ax0.axis[:5], expected_data_start_X)


class TestPhotonCount:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(testfile_photon_count_path, reader="Hamamatsu")

    @classmethod
    def teardown_class(cls):
        del cls.s
        gc.collect()

    def test_data(self):
        expected_data = [0, 0, 0]
        np.testing.assert_allclose(self.s.isig[-3:, 0].data, expected_data)

    def test_metadata(self):
        metadata = self.s.metadata
        assert metadata.General.date == "2018-08-29"
        assert (
            metadata.Acquisition_instrument.Detector.acquisition_mode
            == "photon_counting"
        )
        assert "Grating" not in list(
            self.s.metadata.Acquisition_instrument.Spectrometer.as_dictionary()
        )
        assert self.s.original_metadata.file_type == "bit16"


class TestShading:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(testfile_shading_path, reader="Hamamatsu")

    @classmethod
    def teardown_class(cls):
        del cls.s
        gc.collect()

    def test_metadata(self):
        np.testing.assert_allclose(
            self.s.metadata.Acquisition_instrument.Detector.time_range, 4
        )

    def test_data(self):
        expected_data = [9385, 8354, 7658]
        np.testing.assert_allclose(self.s.isig[:3, 0].data, expected_data)


class TestOther:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(testfile_xaxisother_path, reader="Hamamatsu")

    @classmethod
    def teardown_class(cls):
        del cls.s
        gc.collect()

    def test_xaxisother(self):
        # Test case where ScalingXScalingFile="Other"
        assert self.s.axes_manager.signal_shape == (672, 508)
        assert self.s.axes_manager.navigation_shape == ()
        assert self.s.data.shape == (508, 672)
        assert self.s.axes_manager[0].units == "px"
        assert self.s.axes_manager[0].name == "Uncalibrated X axis"
        assert self.s.axes_manager[1].units == "px"
        assert self.s.axes_manager[1].name == "Vertical CCD Position"
        np.testing.assert_allclose(self.s.axes_manager[1].scale, 1.0, rtol=1e-3)
        np.testing.assert_allclose(
            self.s.axes_manager[1].offset, 0.0, rtol=1e-3, atol=1e-5
        )
        np.testing.assert_allclose(self.s.axes_manager[0].scale, 1.0, rtol=1e-3)
        np.testing.assert_allclose(
            self.s.axes_manager[0].offset, 0.0, rtol=1e-3, atol=1e-5
        )



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging
import os
import zipfile
from pathlib import Path

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
# zarr (because of numcodecs) is only supported on x86_64 machines
zarr = pytest.importorskip("zarr", reason="zarr not installed")

dname = Path(os.path.dirname(__file__)) / "data" / "zspy"


class TestZspy:
    @pytest.fixture
    def signal(self):
        data = np.ones((10, 10, 10, 10))
        s = hs.signals.Signal1D(data)
        return s

    @pytest.mark.parametrize("store_class", [zarr.N5Store, zarr.ZipStore])
    def test_save_store(self, signal, tmp_path, store_class):
        filename = tmp_path / "test_save_store.zspy"
        store = store_class(path=filename)
        signal.save(store)

        if store_class is zarr.ZipStore:
            assert os.path.isfile(filename)
        else:
            assert os.path.isdir(filename)

        store2 = store_class(path=filename)
        signal2 = hs.load(store2)

        np.testing.assert_array_equal(signal2.data, signal.data)

    @pytest.mark.parametrize("close_file", [True, False])
    def test_save_ZipStore_close_file(self, signal, tmp_path, close_file):
        filename = tmp_path / "test_zip_Store.zspy"
        store = zarr.ZipStore(path=filename)
        signal.save(store, close_file=close_file)

        assert os.path.isfile(filename)

        s2 = hs.load(filename)
        np.testing.assert_array_equal(s2.data, signal.data)

    def test_save_ZipStore_mode_warning(self, signal, tmp_path, caplog):
        filename = tmp_path / "test.zspy"
        store = zarr.ZipStore(path=filename)
        signal.save(store)

        with caplog.at_level(logging.WARNING):
            _ = hs.load(filename, mode="r+")
            assert "Specifying `mode` " in caplog.text

    def test_save_wrong_store(self, signal, tmp_path, caplog):
        filename = tmp_path / "testmodels.zspy"
        store = zarr.N5Store(path=filename)
        signal.save(store)

        store2 = zarr.N5Store(path=filename)
        s2 = hs.load(store2)
        np.testing.assert_array_equal(s2.data, signal.data)

        store2 = zarr.NestedDirectoryStore(path=filename)
        with pytest.raises(Exception):
            with caplog.at_level(logging.ERROR):
                _ = hs.load(store2)

    @pytest.mark.parametrize("overwrite", [None, True, False])
    def test_overwrite(self, signal, overwrite, tmp_path):
        filename = tmp_path / "testmodels.zspy"
        signal.save(filename=filename)
        signal2 = signal * 2
        signal2.save(filename=filename, overwrite=overwrite)
        if overwrite is None:
            np.testing.assert_array_equal(signal.data, hs.load(filename).data)
        elif overwrite:
            np.testing.assert_array_equal(signal2.data, hs.load(filename).data)
        else:
            np.testing.assert_array_equal(signal.data, hs.load(filename).data)

    def test_compression_opts(self, tmp_path):
        self.filename = tmp_path / "testfile.zspy"
        from numcodecs import Blosc

        comp = Blosc(cname="zstd", clevel=1, shuffle=Blosc.SHUFFLE)
        hs.signals.BaseSignal([1, 2, 3]).save(self.filename, compressor=comp)
        f = zarr.open(self.filename.__str__(), mode="r+")
        d = f["Experiments/__unnamed__/data"]
        assert d.compressor == comp

    @pytest.mark.parametrize("compressor", (None, "default", "blosc"))
    def test_compression(self, compressor, tmp_path):
        if compressor == "blosc":
            from numcodecs import Blosc

            compressor = Blosc(cname="zstd", clevel=3, shuffle=Blosc.BITSHUFFLE)
        s = hs.signals.Signal1D(np.ones((3, 3)))
        s.save(
            tmp_path / "test_compression.zspy", overwrite=True, compressor=compressor
        )
        _ = hs.load(tmp_path / "test_compression.zspy")


def test_non_valid_zspy(tmp_path, caplog):
    filename = tmp_path / "testfile.zspy"
    data = np.arange(10)

    f = zarr.group(filename)
    f.create_dataset("dataset", data=data)

    with pytest.raises(IOError):
        with caplog.at_level(logging.ERROR):
            _ = hs.load(filename)


@pytest.mark.parametrize(
    "fname",
    [
        "signal1d_10x10-DirectoryStore.zspy",
        "signal1d_10x10-NestedDirectoryStore.zspy",
        "signal1d_10x10-ZipStore.zspy",
    ],
)
def test_read_zspy_saved_with_zarr_v2(fname):
    """Test reading a zspy file saved with zarr v2 and different stores."""
    fname = dname / fname

    s = hs.load(fname)
    assert s.data.shape == (10, 10)
    assert s.axes_manager.signal_shape == (10,)


def test_read_zspy_saved_with_zarr_v2_ragged_markers():
    """
    Test reading a zspy file with ragged markers saved with zarr v2.

    File created with the following code:

    import hyperspy.api as hs
    import numpy as np

    data = np.arange(20*10*10).reshape((20, 10, 10))
    s = hs.signals.Signal2D(data)

    offsets = np.empty(s.axes_manager.navigation_shape, dtype=object)
    for ind in np.ndindex(offsets.shape):
        i = ind[0] / 2
        offsets[ind] = [i, i]
    m = hs.plot.markers.Points(
        offsets=offsets,
        facecolor='orange',
        )

    s.plot()
    s.add_marker(m, permanent=True)
    s.save("signal2d_20x10x10-ragged_markers.zspy", store_type="zip")
    """
    fname = dname / "signal2d_20x10x10-ragged_markers.zspy"

    s = hs.load(fname)
    assert s.data.shape == (20, 10, 10)
    m = s.metadata.Markers.Points
    # The position of the markers increase by 0.5 in each dimension
    # with increase the navigation position
    for i, array in enumerate(m.kwargs["offsets"]):
        np.testing.assert_allclose(array, [i / 2, i / 2])


def test_read_zspy_saved_with_zarr_v2_ragged_markers_unicode():
    """
    Test reading a zspy file with ragged markers and unicode dtype
    saved with zarr v2.

    File created with the following code:

    import hyperspy.api as hs
    import numpy as np

    data = np.ones((5, 10, 10))
    s = hs.signals.Signal2D(data)

    offsets = np.empty(s.axes_manager.navigation_shape, dtype=object)
    texts = np.empty(s.axes_manager.navigation_shape, dtype=object)
    for index in np.ndindex(offsets.shape):
        i = index[0]
        offsets[index] = [i/2, i/2]
        texts[index] = np.array(["a" * (i + 1), "b",][: i + 2])
    m = hs.plot.markers.Texts(
        offsets=offsets,
        texts=texts,
        sizes=3,
        facecolor="black",
    )

    s.add_marker(m, permanent=True)
    s.save("signal2d_20x10x10-ragged_markers_unicode.zspy", store_type="zip")
    """
    fname = dname / "signal2d_20x10x10-ragged_markers_unicode.zspy"

    s = hs.load(fname)
    assert s.data.shape == (5, 10, 10)
    m = s.metadata.Markers.Texts
    assert m.kwargs["offsets"].dtype == object
    assert m.kwargs["texts"].dtype == object
    np.testing.assert_equal(m.kwargs["texts"][0], ["a", "b"])
    np.testing.assert_equal(m.kwargs["texts"][1], ["aa", "b"])


def test_read_zspy_saved_with_zarr_v2_structured_array():
    """
    Test reading a zspy file with structured arrays saved with zarr v2.
    A multidimensional model is good example of a structured array, because
    the parameters map is a structured array.

    import numpy as np
    import hyperspy.api as hs

    # Generate the data and make the spectrum
    data = np.arange(1000, dtype=np.int64).reshape((10, 100))
    s = hs.signals.Signal1D(data)
    s.add_poissonian_noise(random_state=0)

    m = s.create_model()
    line = hs.model.components1D.Expression("a * x + b", name="Affine")
    m.append(line)

    m.multifit()
    m.store()
    s.save("signal1d_10x100-model.zspy", store_type="zip")

    """
    fname = dname / "signal1d_10x100-model.zspy"

    s = hs.load(fname)
    m = s.models.restore("a")

    assert m[0].a.map.dtype == np.dtype(
        [("values", "<f8"), ("std", "<f8"), ("is_set", bool)]
    )


@pytest.mark.parametrize("store_type", ["local", "zip"])
def test_save_store(store_type, tmp_path):
    filename = tmp_path / f"test_save_{store_type}.zspy"

    data = np.ones((10, 10, 10, 10))
    s = hs.signals.Signal1D(data)
    s.save(filename, store_type=store_type)

    if store_type == "zip":
        assert zipfile.is_zipfile(filename)
    else:
        assert filename.is_dir()

    s2 = hs.load(filename)
    np.testing.assert_allclose(s2.data, s.data)


def test_save_store_error(tmp_path):
    # ValueError if store_type is not None and a zarr store
    # is passed to filename
    data = np.ones((10, 10, 10, 10))
    s = hs.signals.Signal1D(data)

    with pytest.raises(ValueError, match="one of 'local' or 'zip'."):
        s.save(tmp_path / "test0.zspy", store_type="unsupported_store_type")

    store = zarr.storage.ZipStore(tmp_path / "test1.zspy")
    with pytest.raises(
        ValueError, match="The `store_type` parameter must be None if a zarr "
    ):
        s.save(filename=store, store_type="zip")



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


from contextlib import contextmanager


@contextmanager
def dummy_context_manager(*args, **kwargs):
    yield


def get_progress_bar_context_manager(show_progressbar):
    """
    Wrapper to `dask.diagnostics.ProgressBar` context manager that
    defer dask import and add argument to enable/disable the progress bar.

    Parameters
    ----------
    show_progressbar : bool
        Whether to show the progress bar.

    Returns
    -------
    context manager
        A context manager that shows a progress bar if requested,
        otherwise it will return a dummy context manager.
    """
    if show_progressbar:
        from dask.diagnostics import ProgressBar

        return ProgressBar
    else:
        return dummy_context_manager



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO.  If not, see <https://www.gnu.org/licenses/#GPL>.

from pathlib import Path

import numpy as np
import pytest

from rsciio.impulse._api import ImpulseCSV, invalid_file_error, invalid_filenaming_error

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")


TEST_DATA_PATH = Path(__file__).parent / "data" / "impulse"


# Load a synchronized data log file


def test_read_sync_file():
    filename = TEST_DATA_PATH / "StubExperiment_Synchronized data.csv"
    s = hs.load(filename, reader="impulse")
    assert len(s) == 13
    assert s[0].metadata.General.title == "Temperature Measured (degC)"
    assert s[0].metadata.Signal.quantity == "degC"
    assert s[1].metadata.General.title == "MixValve"
    assert s[1].metadata.Signal.quantity == ""
    assert s[2].metadata.General.title == "Pin Measured"
    assert s[2].metadata.Signal.quantity == ""
    assert s[3].metadata.General.title == "Pout Measured"
    assert s[3].metadata.Signal.quantity == ""
    assert s[4].metadata.General.title == "Pnr Measured"
    assert s[4].metadata.Signal.quantity == ""
    assert s[5].metadata.General.title == "Fnr"
    assert s[5].metadata.Signal.quantity == ""
    assert s[6].metadata.General.title == "% Gas1 Measured"
    assert s[6].metadata.Signal.quantity == ""
    assert s[7].metadata.General.title == "% Gas2 Measured"
    assert s[7].metadata.Signal.quantity == ""
    assert s[8].metadata.General.title == "% Gas3 Measured"
    assert s[8].metadata.Signal.quantity == ""
    assert s[9].metadata.General.title == "Channel#1"
    assert s[9].metadata.Signal.quantity == ""
    assert s[10].metadata.General.title == "Channel#2"
    assert s[10].metadata.Signal.quantity == ""
    assert s[11].metadata.General.title == "Channel#3"
    assert s[11].metadata.Signal.quantity == ""
    assert s[12].metadata.General.title == "Channel#4"
    assert s[12].metadata.Signal.quantity == ""


class testSyncFile:
    def setup_method(self, method):
        filename = TEST_DATA_PATH / "StubExperiment_Synchronized data.csv"
        self.s_list = hs.load(filename, reader="impulse")

    def test_read_metadata(self):
        for s in self.s_list:
            assert s.metadata.General.date == "06-08-2021"
            assert s.metadata.General.time == "13:40:19"
            assert s.axes_manager[0].units == "s"
            np.testing.assert_allclose(
                s.axes_manager[0].scale, 0.5114766839378239, atol=1e-5
            )
            assert s.axes_manager[0].offset == 0

    def test_read_original_metadata(self):
        om = self.s_list[0].original_metadata.Impulse_header
        assert om.Calibration_value == 0.25
        assert om.Description == "Experiment with stub data"
        assert om.Experiment_date == "06-08-2021"
        assert om.Experiment_duration == "00.00:06:45.416"
        assert om.Experiment_time == "13:40:19"
        assert om.Gas1 == "H2"
        assert om.Gas2 == "CH4"
        assert om.Gas3 == "Ar"
        assert om.MassSpec_File_Location == ""
        assert om.MassSpec_channel_1 == "28 (Nitrogen)"
        assert om.MassSpec_channel_2 == "32 (Oxygen)"
        assert om.MassSpec_channel_3 == "18 (Water)"
        assert om.MassSpec_channel_4 == "44 (Carbon dioxide)"
        assert om.Max_temperature == 1100
        assert om.Notes == [
            "13:46:46: Live note one",
            "13:46:52: Live note two",
            "13:46:56: Live note three",
        ]
        assert om.Room_temperature == 21
        assert om.Sample == "Gold"
        assert om.System_version == "G+"

    def test_read_data(self):
        expected_data = np.load(TEST_DATA_PATH / "StubExperiment_Synchronized data.npy")
        np.testing.assert_allclose(self.s_list.T, expected_data)


class testSyncFileCSVreader:
    def setup_method(self, method):
        self.filename = TEST_DATA_PATH / "StubExperiment_Synchronized data.csv"
        self.isf = ImpulseCSV(self.filename)

    def test_read_column_name(self):
        assert self.isf.column_name == [
            "Temperature Measured",
            "MixValve",
            "Pin Measured",
            "Pout Measured",
            "Pnr Measured",
            "Fnr",
            "% Gas1 Measured",
            "% Gas2 Measured",
            "% Gas3 Measured",
            "Channel#1",
            "Channel#2",
            "Channel#3",
            "Channel#4",
        ]

    def test_read_data(self):
        dicts = (
            self.isf._data_dictionary[key] for key in self.isf.logged_quantity_name_list
        )
        data = np.vstack(list(dicts))
        expected_data = np.load(TEST_DATA_PATH / "StubExperiment_Synchronized data.npy")
        np.testing.assert_allclose(data.T, expected_data)


# Loading a random csv file


def test_loading_random_csv_file():
    filename = TEST_DATA_PATH / "random_csv_file.csv"
    with pytest.raises(IOError) as cm:
        ImpulseCSV(filename)
        cm.match(invalid_file_error)


# Loading a csv file with an incorrect filename


def test_loading_invalid_impulse_filename():
    filename = TEST_DATA_PATH / "changed_file_name.csv"
    with pytest.raises(IOError) as cm:
        hs.load(filename, reader="impulse")
        cm.match(invalid_filenaming_error)


# Test raw data file


class testRawFile:
    def setup_method(self, method):
        filename = TEST_DATA_PATH / "StubExperiment_Heat raw.csv"
        self.s_list = hs.load(filename, reader="impulse")

    def test_read_metadata(self):
        for s in self.s_list:
            assert s.metadata.General.date == "06-08-2021"
            assert s.metadata.General.time == "13:40:19"
            assert s.axes_manager[0].units == "s"
            np.testing.assert_allclose(
                s.axes_manager[0].scale, 0.5114766839378239, atol=1e-5
            )
            assert s.axes_manager[0].offset == 0

    def test_read_original_metadata(self):
        om = self.s_list[0].original_metadata.Impulse_header
        assert om.Calibration_value == 0.25
        assert om.Description == "Experiment with stub data"
        assert om.Experiment_date == "06-08-2021"
        assert om.Experiment_duration == "00.00:06:45.416"
        assert om.Experiment_time == "13:40:19"
        assert om.Gas1 == "H2"
        assert om.Gas2 == "CH4"
        assert om.Gas3 == "Ar"
        assert om.MassSpec_File_Location == ""
        assert om.MassSpec_channel_1 == "28 (Nitrogen)"
        assert om.MassSpec_channel_2 == "32 (Oxygen)"
        assert om.MassSpec_channel_3 == "18 (Water)"
        assert om.MassSpec_channel_4 == "44 (Carbon dioxide)"
        assert om.Max_temperature == 1100
        assert om.Notes == [
            "13:46:46: Live note one",
            "13:46:52: Live note two",
            "13:46:56: Live note three",
        ]
        assert om.Room_temperature == 21
        assert om.Sample == "Gold"
        assert om.System_version == "G+"


class testRawFileCSVreader:
    def setup_method(self, method):
        self.filename = TEST_DATA_PATH / "StubExperiment_Heat raw.csv"
        self.isf = ImpulseCSV(self.filename)

    def test_read_column_name(self):
        assert self.isf.column_name == ["Temperature Measured"]

    def test_read_start_datetime(self):
        assert self.isf.start_datetime == "2021-08-06T13:40:19.000000"



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.
"""Utility functions for XML handling."""

import re
import xml.etree.ElementTree as ET
from ast import literal_eval
from collections import defaultdict

from rsciio.utils._dictionary import DTBox

__all__ = [
    "XmlToDict",
    "convert_xml_to_dict",
    "sanitize_msxml_float",
    "xml2dtb",
]


def __dir__():
    return sorted(__all__)


# MSXML sanitization ###
# re pattern with two capturing groups with comma in between;
# firstgroup looks for numeric value after <tag> (the '>' char) with or
# without minus sign, second group looks for numeric value with following
# closing <\tag> (the '<' char); '([Ee]-?\d*)' part (optionally a third group)
# checks for scientific notation (e.g. 8,843E-7 -> 'E-7');
# compiled pattern is binary, as raw xml string is binary string.:
_fix_dec_patterns = re.compile(b"(>-?\\d+),(\\d*([Ee]-?\\d*)?<)")


def sanitize_msxml_float(xml_b_string):
    """
    Replace comma with dot in floating point numbers in given xml
    raw string.

    Parameters
    ----------
    xml_b_string : str
        Raw binary string representing the xml to be parsed.

    Returns
    -------
    str
        Binary string with commas used as decimal marks replaced with dots
        to adhere to XML standard.

    Notes
    -----
    What, why, how? In case OEM software runs on MS Windows and directly
    uses system built-in MSXML lib, which does not comply with XML standards,
    and OS is set to locale of some country with weird and illogical
    preferences of using comma as decimal separation;
    Software in conjunction of these above listed conditions can produce
    not-interoperable XML, which leads to wrong interpretation of context.
    This sanitizer searches and corrects that - it should be used before
    being fed to .fromstring of element tree.
    """
    return _fix_dec_patterns.sub(b"\\1.\\2", xml_b_string)


class XmlToDict:
    """
    Customisable XML to python dict and list based Hierarchical tree
    translator.

    Parameters
    ----------
    dub_attr_pre_str : str
        String to be prepended to attribute name when creating
        dictionary tree if children element with same name is used.
        Default is "@".
    dub_text_str : str (default: "#text")
        String to use as key in case element contains text and children tag.
        Default "#text".
    tags_to_flatten : None, str or list of str
        Define tag names which should be flattened/skipped,
        placing children of such tag one level shallower in constructed
        python structure.
        It is useful when OEM generated XML are not human designed,
        but machine/programming language/framework generated
        and painfully verbose. See example below.
        Default is None, which means no tags are flattened.
    interchild_text_parsing : str
        Must be one of ("skip", "first", "cat", "list").
        This considers the  behaviour when both .text and children tags are
        presented under same element tree node:

        - "skip" - will not try to retrieve any .text values from such node.
        - "first" - only string under .text attribute will be returned.
        - "cat" - return concatenated string from .text of node and .tail's
          of children nodes.
        - "list" - similar to "cat", but return the result in list
          without concatenation.

        Default is "first", which is the most common case.

    Examples
    --------
    Consider such redundant tree structure:

    .. code-block::

        DetectorHeader
        |-ClassInstances
            |-ClassInstance
            |-Type
            |-Window
            ...

    It can be sanitized/simplified by setting tags_to_flatten keyword
    with ["ClassInstances", "ClassInstance"] to eliminate redundant
    levels of tree with such tag names:

    .. code-block::

        DetectorHeader
        |-Type
        |-Window
        ...

    Produced dict/list structures are then good enough to be
    returned as part of original metadata without making any more
    copies.

    Setup the parser:

    >>> from rsciio.utils.xml import XmlToDict
    >>> xml_to_dict = XmlToDict(
    ...     pre_str_dub_attr="XmlClass",
    ...     tags_to_flatten=[
    ...         "ClassInstance", "ChildrenClassInstance", "JustAnotherRedundantTag"
    ...     ]
    ... )

    Use parser:

    >>> pytree = xml_to_dict.dictionarize(etree_node)

    """

    def __init__(
        self,
        dub_attr_pre_str="@",
        dub_text_str="#value",
        tags_to_flatten=None,
        interchild_text_parsing="first",
    ):
        if tags_to_flatten is None:
            tags_to_flatten = []
        if type(tags_to_flatten) not in [str, list]:
            raise ValueError(
                "tags_to_flatten keyword accepts string or list of strings"
            )
        if not isinstance(dub_attr_pre_str, str):
            raise ValueError("dub_attr_pre_str should be of string type")
        if not isinstance(dub_text_str, str):
            raise ValueError("dub_text_str should be of string type")
        if isinstance(tags_to_flatten, str):
            tags_to_flatten = [tags_to_flatten]
        if interchild_text_parsing not in ("skip", "first", "cat", "list"):
            raise ValueError(
                "interleaved_text_parsing should be set to the one of: "
                "('skip', 'first', 'cat', 'list')"
            )
        self.tags_to_flatten = tags_to_flatten
        self.dub_attr_pre_str = dub_attr_pre_str
        self.dub_text_str = dub_text_str
        self.poor_text_mode = interchild_text_parsing

    @staticmethod
    def eval(string):
        """
        Interpret any string and return casted to appropriate
        dtype python object.

        Parameters
        ----------
        string : str
            String to be interpreted.

        Returns
        -------
        string
            Interpreted string.

        Notes
        -----
        If this does not return desired type, consider subclassing
        and reimplementing this method like this:

        .. code-block:: python

            class SubclassedXmlToDict(XmlToDict):
                @staticmethod
                def eval(string):
                    if condition check to catch the case
                    ...
                    elif
                    ...
                    else:
                        return XmlToDict.eval(string)
        """
        try:
            return literal_eval(string)
        except (ValueError, SyntaxError):
            # SyntaxError due to:
            # literal_eval have problems with strings like this '8842_80'
            return string

    def dictionarize(self, et_node):
        """
        Take etree XML node and return its conversion into
        pythonic dict/list representation of that XML tree
        with some sanitization.

        Parameters
        ----------
        et_node : xml.etree.ElementTree.Element
            XML node to be converted.

        Returns
        -------
        dict
            Dictionary representation of the XML node.
        """
        d_node = {et_node.tag: {} if et_node.attrib else None}
        children = list(et_node)
        if children:
            dd_node = defaultdict(list)
            for dc_node in map(self.dictionarize, children):
                for key, val in dc_node.items():
                    dd_node[key].append(val)
            d_node = {
                et_node.tag: {
                    key: self.eval(val[0]) if len(val) == 1 else val
                    for key, val in dd_node.items()
                }
            }
        if et_node.attrib:
            d_node[et_node.tag].update(
                (self.dub_attr_pre_str + key if children else key, self.eval(val))
                for key, val in et_node.attrib.items()
            )
        if et_node.text:
            text = et_node.text.strip()
            if text:
                if not et_node.attrib and not children:
                    d_node[et_node.tag] = self.eval(text)
                elif children:
                    if self.poor_text_mode == "first":
                        d_node[et_node.tag][self.dub_text_str] = self.eval(text)
                    elif self.poor_text_mode in ("cat", "list"):
                        tails = [
                            str(c.tail if c.tail is not None else "").strip()
                            for c in children
                        ]
                        if any(tails):
                            inter_pieces = [text]
                            inter_pieces.extend(tails)
                            if self.poor_text_mode == "cat":
                                inter_pieces = "".join(inter_pieces)
                            d_node[et_node.tag]["#interchild_text"] = inter_pieces
                        else:
                            d_node[et_node.tag][self.dub_text_str] = self.eval(text)
                elif et_node.attrib:
                    d_node[et_node.tag][self.dub_text_str] = self.eval(text)
        for tag in self.tags_to_flatten:
            if tag in d_node:
                return d_node[tag]
        return d_node


def xml2dtb(et, dictree):
    """
    Convert XML ElementTree node to DTBox object.
    This is a recursive function that traverses the XML tree
    and populates the DTBox object with the data from the XML node.

    Parameters
    ----------
    et : xml.etree.ElementTree.Element
        XML node to be converted.
    dictree : DTBox
        Box object to be populated.

    """
    if et.text:
        dictree.set_item(et.tag, et.text)
        return
    else:
        dictree.add_node(et.tag)
        if et.attrib:
            dictree[et.tag].merge_update(et.attrib)
        for child in et:
            xml2dtb(child, dictree[et.tag])


def convert_xml_to_dict(xml_object):
    """
    Convert XML object to a DTBox object.

    Parameters
    ----------
    xml_object : str or xml.etree.ElementTree.Element
        XML object to be converted. It can be a string or an ElementTree node.

    Returns
    -------
    DTBox
        A DTBox object containing the converted XML data.
    """
    if isinstance(xml_object, str):
        xml_object = ET.fromstring(xml_object)
    op = DTBox(box_dots=True)
    xml2dtb(xml_object, op)
    return op
